from typing import Callable

from twisted.web.server import Request

from canonical.txapi.server.schema import Schema, Unicode

from canonical.landscape.api.helpers import EMPTY_BODY, with_schema
from canonical.landscape.api.jwt_auth import (
    encode_auth_token,
    get_secret_token,
    login_required,
)
from canonical.landscape.api.principal import Principal
from canonical.landscape.api.types import ApiResponse
from canonical.landscape.model.main.person import get_person_by_email, Person
from canonical.landscape.model.main.account import get_account_by_name
from canonical.landscape.security.authentication import LandscapePAMAuthentication
from canonical.pam import pam


INVALID_LOGIN_BODY = {
    "error": "InvalidLoginError",
    "message": "email, account, or password is incorrect.",
}

login_schema = Schema(
    name="Login",
    parameters=[
        Unicode("email"),
        Unicode("password"),
        Unicode("account", optional=True),
    ],
)


def database_authenticator(identity: str, password: str) -> Person | None:
    """Authenticates `identity` against their stored database password."""
    person = get_person_by_email(identity)

    if person and person.check_password(password):
        return person

    return None


def pam_authenticator(identity: str, password: str) -> Person | None:
    """Authenticates `identity` via a PAM-module. This could be LDAP or Active Directory."""
    person, pam_identity = LandscapePAMAuthentication.get_person_identity(identity)

    if person and pam.authenticate(pam_identity, password, "landscape"):
        return person

    return None


# Collection of functions used to authenticate an identity against a password.
# Authenticators return a `Person` for successful authentication, `None` otherwise.
AUTHENTICATORS: tuple[Callable[[str, str], Person | None], ...] = (
    database_authenticator,
    pam_authenticator,
)


@with_schema(login_schema)
async def login_post_handler(request: Request, validated_data: dict) -> ApiResponse:
    """RequestHandler that authenticates a user and provides them with a JWT."""
    email = validated_data["email"]
    password = validated_data["password"]
    account_name = validated_data.get("account")

    for authenticator in AUTHENTICATORS:
        if person := authenticator(email, password):
            break
    else:
        return INVALID_LOGIN_BODY, 400

    default_account = person.get_default_account()

    if account_name:
        login_account = get_account_by_name(account_name)

        if login_account is None or person not in login_account.people:
            return INVALID_LOGIN_BODY, 400
    else:
        login_account = default_account

    accounts = []
    for account in person.accounts:
        account_dict = {
            "title": account.title,
            "name": account.name,
        }

        if account is default_account:
            account_dict["default"] = True

        accounts.append(account_dict)

    return {
        "token": encode_auth_token(person, login_account, get_secret_token()),
        "email": email,
        "name": person.name,
        "accounts": accounts,
        "current_account": login_account.name,
    }, 200


update_password_schema = Schema(
    name="UpdatePassword",
    parameters=[
        Unicode("password"),
        # max and min per c/l/model/main/interfaces.py
        Unicode("new_password", min=3, max=30),
    ],
)


@login_required
@with_schema(update_password_schema)
async def password_put_handler(
    request: Request, principal: Principal, validated_data: dict
) -> ApiResponse:
    """Changes the caller's password to a new one.

    N.B.: this does not invalidate any already-issued JWTs.
    """
    person = principal.person
    current_password = validated_data["password"]

    if not person.check_password(current_password):
        return {
            "error": "IncorrectPasswordError",
            "message": "Incorrect password.",
        }, 400

    person.set_password(validated_data["new_password"])

    return EMPTY_BODY, 204
