# Copyright 2005-2024 Canonical Limited. All rights reserved.

"""
Computers
=========

The methods available here give the ability to retrieve computers and to do
basic operations on them, like tagging.
"""
import typing
import json
from collections import defaultdict
from typing import Sequence, Iterable, Dict, Optional

from landscape.lib.tag import is_valid_tag

from storm.expr import And, Desc, Join, Not

from twisted.web.server import Request

from canonical.search.engine import SearchCriteriaError

from canonical.txapi.server.call import Call
from canonical.txapi.server.method import method
from canonical.txapi.server.schema import (
    Bool,
    Integer,
    List,
    PositiveInteger,
    Schema,
    Unicode,
)

from canonical.lib.text import make_secure_id
from canonical.lib.timestamp import parse_datetime

from canonical.landscape.api import RESULT_OFFSET, RESULT_LIMIT
from canonical.landscape.api.error import (
    CloudInitEncodingError,
    ComputerProcessNotFoundError,
    InsufficientLicensesError,
    InvalidParameterCombinationError,
    InvalidQueryError,
    InvalidTagError,
    InvalidTimeError,
    InvalidTitleError,
    MissingParameterError,
    NotChildComputerError,
    NotWSLHostError,
    SecretNotFoundError,
    SecretsAuthenticationError,
    SecretsSealedError,
    UnknownAccessGroupError,
    UnknownComputerError,
    UnauthorisedError,
    UnknownRelationshipError,
)
from canonical.landscape.api.helpers import (
    ApiRequestError,
    ApiResponse,
    Data,
    add_search_query,
    check_person_permission,
    paginated,
    with_computer,
    with_schema,
)
from canonical.landscape.api.jwt_auth import login_required
from canonical.landscape.api.methods import AccountMethod, search_schema
from canonical.landscape.api.principal import Principal
from canonical.landscape.api.routes import AccountRoute, SelectionRoute
from canonical.landscape.api.schema import Array, Mapping, UnicodeTitle

from canonical.landscape.model.account.store import get_account_stores
from canonical.landscape.model.activity.activity import Activity
from canonical.landscape.model.activity.collection import ActivityChildCollection
from canonical.landscape.model.activity.group import (
    SIGNAL_PROCESSES_SUMMARY,
    ActivityGroup,
)
from canonical.landscape.model.alert.collection import RuleCollection
from canonical.landscape.model.alert.computer import ComputerRuleInfo
from canonical.landscape.model.alert.rule import LEGACY_ALERTS, RuleState
from canonical.landscape.model.computer.collection import (
    AccountComputerTwinCollection,
    ComputerCollection,
)
from canonical.landscape.model.computer.management import ComputerManagement
from canonical.landscape.model.computer.search import ComputerSearch
from canonical.landscape.model.computer.twin import AccountComputerTwin
from canonical.landscape.model.cloud.instance import CloudRegistrationOTP
from canonical.landscape.model.entity.helper import (
    get_access_context_ids_by_entity_type,
)
from canonical.landscape.model.main.account import (
    DEFAULT_ACCESS_CONTEXT_NAME,
    Account,
    get_account_by_id,
)
from canonical.landscape.model.main.computer import (
    Computer,
    ComputerRelationship,
    PendingComputerToAccept,
    add_tags_to_computers,
    get_computer_by_id,
    remove_tags_from_computers,
)
from canonical.landscape.model.main.distribution import (
    WSL_HOST_CODE_NAME,
    WSL_INSTANCE_NAMES,
    Distribution,
    get_distribution,
)
from canonical.landscape.model.main.errors import (
    AcceptanceUnauthorized,
    UnknownAccessContextsError,
)
from canonical.landscape.model.main.license import LicenseSeatError
from canonical.landscape.model.main.person import Person, get_person_by_id
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.tag import (
    InvalidTagError as InvalidTagModelError,
    get_tags,
    get_tag_names,
)
from canonical.landscape.model.package.alerts import PackageUpgradesAlert
from canonical.landscape.model.package.search import PackageSearch
from canonical.landscape.model.resource.activeprocessinfo import (
    SIGNAL_KILL,
    SIGNAL_TERM,
)
from canonical.landscape.model.resource.activeprocessinfo import ActiveProcess
from canonical.landscape.model.resource.annotation import (
    AnnotationManagement,
    InvalidAnnotationKeyError,
)
from canonical.landscape.model.resource.hardware import HardwareInfoManagement
from canonical.landscape.model.resource.hardware_info_parser import HardwareInfoParser
from canonical.landscape.model.resource.network import AccountNetworkInfo
from canonical.landscape.model.selection.selection import ComputerSelection
from canonical.landscape.service.activity.get import get_children_for_query
from canonical.landscape.service.computer.accesscontextchange import (
    ComputerAccessContextChange,
)
from canonical.landscape.service.computer.remove import (
    RemoveComputers as RemoveComputersOperation,
)
from canonical.landscape.service.collection.errors import (
    InvalidQueryError as InvalidSearchError,
    UnknownEntityIDsError,
)
from canonical.landscape.service.security import PersonPrincipal


__all__ = [
    "GetComputers",
    "AddAnnotationToComputers",
    "RemoveAnnotationFromComputers",
    "AddTagsToComputers",
    "RemoveTagsFromComputers",
    "ChangeComputersAccessGroup",
    "RemoveComputers",
    "GetPendingComputers",
    "AcceptPendingComputers",
    "RejectPendingComputers",
    "CreateCloudOtps",
    "RebootComputers",
    "ShutdownComputers",
    "RenameComputers",
    "GetWSLHosts",
    "DeleteChildComputers",
    "StartChildComputers",
    "StopChildComputers",
    "GetComputerProcesses",
    "TerminateComputerProcesses",
    "KillComputerProcesses",
    "signal_computer_proc_post_handler",
    "computers_activities_list_handler",
    "computers_get_handler",
    "computers_put_handler",
    "computers_list_handler",
    "computer_processes_list_handler",
    "wsl_hosts_list_handler",
]


computers_extra_parameters = [
    Bool(
        "with_network",
        optional=True,
        default=False,
        doc=(
            "If true, include the details of all network devices "
            "attached to the computer."
        ),
    ),
    Bool(
        "with_hardware",
        optional=True,
        default=False,
        doc="If true, include the details of all known hardware information.",
    ),
    Bool(
        "with_annotations",
        optional=True,
        default=False,
        doc=(
            "If true, include the details of all custom annotation "
            "information known."
        ),
    ),
    Bool(
        "with_grouped_hardware",
        optional=True,
        default=False,
        doc=(
            "If true, include the details of all known hardware information grouped by "
            "device category."
        ),
    ),
]


@method
class GetComputers(AccountMethod):
    """
    Get a list of computers associated with the account used for
    authentication.

    .. _computer-query:

    - **query**: A query string with space separated tokens used to filter the
      returned computers. Words provided as search parameters are treated as
      keywords, matching the hostname or the computer title. Selector prefixes
      can be used to further customize the search:

    - *tag*: search for computers with the specified tag.
    - *distribution*: search for computers running a specific Ubuntu
      release (can be code name like 'lucid' or version number like
      '10.04').
    - *hostname*: search for computers with the exact hostname.
    - *title*: search for computers with the exact title.
    - *alert*: search for computers with a named alert being active, alerts
      can be one of "package-upgrades", "security-upgrades",
      "package-profiles", "package-reporter",  "computer-offline",
      "computer-reboot".
    - *access-group*: search for computers that belong to the access group with
      the specified name.
    - *id*: select the specified numeric computer id.
    - *mac*: search for computers with the specified MAC address, which can
      be a substring of the full address.
    - *ip*: search for computers with the specified IP address, which can be
      a substring of the full address. No network classing is done.
    - *search*: select computers based on the result of the named search.
    - *needs:reboot*: search for computers that have the reboot required flag
      set.  Note that with this particular criteria, the only possible value
      for it is the text "reboot".
    - *license-id*: search for computers licensed to the specified license-id.
    - *needs:license* OR *license-id:none*: search for computers that do not
      have a Landscape license, and, as a result, are not managed.
    - *annotation*: search for computers which define the specified annotation
      key. Optionally specify annotation:<key>:<string> which will only return
      computers whose key matches and value also contains the <string>
      specified.
    - *OR*: search for computers matching term A or term B. OR must be in
      all caps.
    - *NOT*: search for computers not matching the next term. NOT must be in
      all caps.

    If values following a prefix contains spaces or non-ASCII characters, it
    must be quoted.

    These prefixes can be mixed and matched with keywords.  For example, the
    following query would match computers with 'appserver' in their title or
    hostname, either with the tag 'server' or running the 'hardy' release of
    Ubuntu::

      appserver tag:server OR distribution:hardy

    It also supports the following optional arguments:

    - **limit**: The maximum number of results returned by the method. It
      defaults to 1000.
    - **offset**: The offset inside the list of results.
    - **with_network**: If true, include the details of all network devices
      attached to the computer.
    - **with_hardware**: If true, include the details of all hardware
      information known.
    - **with_annotations**: If true, include the details of all custom
      annotation information known.

    For example, the following request looks for all computers with the tag
    'server', includes their network devices, and with a limit of 20 machines::

        ?action=GetComputers&query=tag:server&limit=20&with_network=true

    The method returns a JSON serialized list of computers::

         [{"access_group": "global",
           "comment": "",
           "distribution": "12.04",
           "hostname": "a_comp.example.com",
           "id": 12345,
           "last_ping_time": None,
           "last_exchange_time": "2011-06-3017:59Z",
           "title": "A Computer",
           "reboot_required_flag": False,
           "tags": ["server"],
           "total_memory": None,
           "total_swap": None,
           "secrets_name": None,
           "network_devices": [{"broadcast_address": "192.168.1.255",
                                "interface": "eth0",
                                "ip_address": "192.168.1.2",
                                "mac_address": "00:1e:c9:6c:b8:de",
                                "netmask": "255.255.255.0"}]}]

    """

    requires = "ViewAccount"

    schema = search_schema.extend(
        name="GetComputers",
        parameters=computers_extra_parameters,
        doc="Get a list of computers associated with the account.",
    )

    route = SelectionRoute(permission="ViewComputer", error_if_empty=False)

    def query(self, context, call):
        """
        Perform a query to get L{Computer}s associated with the
        L{Account}, using the account attached to the credentials making
        the call.
        """
        computer_data = {
            computer.id: computer.get_state() for computer in context.get_selected()
        }
        computer_ids = list(computer_data.keys())

        if call.args.with_network:
            info = AccountNetworkInfo(context.account)
            computer_devices = info.get_devices_for_computer_ids(computer_ids)

            for computer_id, devices in computer_devices.items():
                # TODO: See if we really need to sort the network devices.
                computer_data[computer_id]["network_devices"] = sorted(
                    (device.get_state() for device in devices),
                    key=lambda x: x["interface"],
                )

        if call.args.with_hardware:
            management = HardwareInfoManagement(context.account)
            hardware = management.get_hardware_for_computer_ids(computer_ids)

            for computer_id, data in hardware.items():
                computer_data[computer_id]["hardware"] = data

        if call.args.with_annotations:
            person = get_person_by_id(call.principal.id)
            management = AnnotationManagement(context.account, person)
            annotations = management.get_annotation(computer_ids)

            for computer_id, data in annotations.items():
                computer_data[computer_id]["annotations"] = data

        if call.args.with_grouped_hardware:
            management = HardwareInfoManagement(context.account)
            hardware = management.get_hardware_for_computer_ids(computer_ids)

            for computer_id, data in hardware.items():
                parser = HardwareInfoParser(computer_id, data)
                hardware = parser.hardware

                computer_data[computer_id]["grouped_hardware"] = hardware

        return list(computer_data.values())


ComputersListSchema = GetComputers.schema.extend(
    name="ComputersList",
    parameters=[
        Bool(
            "root_only",
            optional=True,
            default=True,
            doc=(
                "Whether or not to include only the root member of a computer "
                "family tree. Defaults to True. If False, all computers will "
                "be included."
            ),
        ),
        Bool(
            "wsl_only",
            optional=True,
            default=False,
            doc=(
                "Whether or not to include only computers that are WSL instances. "
                "Defaults to False."
            ),
        ),
        Bool(
            "with_alerts",
            optional=True,
            default=False,
            doc=(
                "If true, includes alert information in each computer object if that "
                "alert is active."
            ),
        ),
        Bool(
            "with_upgrades",
            optional=True,
            default=False,
            doc=(
                "If true, includes how many regular and security upgrades that "
                "computer has"
            ),
        ),
    ],
)


ComputersGetSchema = Schema(
    name="GetComputersHandler",
    parameters=computers_extra_parameters,
    doc="Get a computer by id",
)


ComputersPutSchema = Schema(
    name="PutComputersHandler",
    parameters=[
        UnicodeTitle("title", optional=True, doc="The display name for the instance."),
        UnicodeTitle(
            "access_group",
            optional=True,
            doc="The access group the instance belongs to.",
        ),
        Array(
            "tags", Unicode(), optional=True, doc="Tags to associate with the instance."
        ),
        Unicode(
            "comment",
            optional=True,
            doc="A freeform comment to associate with the instance.",
        ),
    ],
    doc="Update an instance by ID.",
)


def get_rule_collection(account_id):
    # Note sharding is not supported
    _, account_store = list(get_account_stores())[0]
    return RuleCollection(account_store, account_id=account_id)


def get_alerts_by_computer(account_id, computer_ids) -> Dict[int, Sequence[Dict]]:
    """Returns a list of alerts by computer id"""
    computer_alerts_data = defaultdict(list)
    collection = get_rule_collection(account_id)
    for rule in collection.find():
        alert = rule.info
        if not isinstance(alert, ComputerRuleInfo):
            continue  # This alert is not related to a computer
        alert_type = alert.get_name()
        if alert_type in LEGACY_ALERTS:
            continue
        result = alert.get_states(active=True, computer_ids=computer_ids)
        alerted_computer_ids = list(result.values(RuleState.computer_id))
        for computer_id in alerted_computer_ids:
            alert_data = {
                "type": alert_type,
                "summary": alert.get_short_summary(),
                "severity": alert.severity,
            }
            computer_alerts_data[computer_id].append(alert_data)
    return computer_alerts_data


def get_num_upgrades_by_computer(
    account, person, computer_ids
) -> Dict[int, Dict[str, int]]:
    """Returns a count of regular and security upgrades by computer id"""
    computer_upgrades_data = {}
    for computer_id in computer_ids:  # Initialize upgrade data
        computer_upgrades_data[computer_id] = {"regular": 0, "security": 0}
    collection = get_rule_collection(account.id)
    rule = collection.with_info_classes([PackageUpgradesAlert]).find().one()
    if not rule:
        return {}
    result = rule.info.get_states(active=True, computer_ids=computer_ids)
    upgrades_computer_ids = list(result.values(RuleState.computer_id))
    if not upgrades_computer_ids:
        return {}
    selection = ComputerSelection(account, person)
    selection.select_computers_by_id(upgrades_computer_ids)
    search = PackageSearch(selection)
    result = search.get_relations_by_package_name(upgrades=True, limit=None)
    packages = result["packages"]
    for relation in result["relations"]:
        for rel_info in relation["relations"]:
            upgrade_id = rel_info["upgrade"]
            upgrade = packages[upgrade_id]
            usn = upgrade.usn
            relation_computer_ids = rel_info["computers"]
            for rid in relation_computer_ids:
                upgrade_data = computer_upgrades_data[rid]
                if upgrade:
                    upgrade_data["regular"] += 1
                if usn:
                    upgrade_data["security"] += 1
    return computer_upgrades_data


@login_required
@with_schema(ComputersListSchema)
@paginated
async def computers_list_handler(
    request,
    principal,
    validated_data,
    limit=None,
    offset=None,
):
    permission = "landscape.ViewComputer"

    collection = ComputerCollection(
        account_id=principal.account_id,
        person_id=principal.id,
        permission=permission,
    )

    if "query" in validated_data and validated_data["query"]:
        try:
            collection = collection.with_search_criteria(validated_data["query"])
        except SearchCriteriaError as error:
            return {"error": "InvalidQueryError", "message": str(error)}, 400

    result = collection.find()

    root_only = validated_data["root_only"]
    wsl_only = validated_data["wsl_only"]

    if root_only and wsl_only:
        return {
            "error": "InvalidParameterCombinationError",
            "message": "Cannot have root_only and wsl_only both be True.",
        }, 400

    if root_only:
        store = get_main_store()
        child_ids = store.find(ComputerRelationship.computer_id)
        result = result.find(Not(Computer.id.is_in(child_ids)))

    if wsl_only:
        store = get_main_store()
        windows_distribution_id = store.find(
            Distribution.id, Distribution.code_name == WSL_HOST_CODE_NAME
        ).one()
        if not windows_distribution_id:
            return {
                "error": "DistributionNotFoundError",
                "message": f"Distribution with code_name '{WSL_HOST_CODE_NAME}' not found.",
            }, 404
        windows_ids = store.find(
            Computer.id, Computer.distribution_id == windows_distribution_id
        )
        wsl_ids = store.find(
            ComputerRelationship.computer_id,
            ComputerRelationship.parent_id.is_in(windows_ids),
        )
        result = result.find(Computer.id.is_in(wsl_ids))

    count = result.count()

    result.order_by(Computer.title)
    result.config(offset=offset, limit=limit)
    computers = list(result)

    account = get_account_by_id(principal.account_id)
    person = get_person_by_id(principal.id)

    computer_data = {computer.id: computer.serialize() for computer in computers}
    computer_ids = list(computer_data.keys())

    extra_data = _get_extra_computer_info(validated_data, account, person, computer_ids)

    for computer_id, extra_data_for_computer in extra_data.items():
        computer_data[computer_id].update(extra_data_for_computer)

    if validated_data["with_alerts"]:
        alerts_data = get_alerts_by_computer(principal.account_id, computer_ids)
        for computer_id, alerts in alerts_data.items():
            computer_data[computer_id]["alerts"] = alerts

    if validated_data["with_upgrades"]:
        upgrades_data = get_num_upgrades_by_computer(
            principal.account, principal.person, computer_ids
        )
        for computer_id, upgrades in upgrades_data.items():
            computer_data[computer_id]["upgrades"] = upgrades

    return {
        "count": count,
        "results": list(computer_data.values()),
    }, 200


def _get_extra_computer_info(
    validated_data: dict,
    account: Account,
    person: Person,
    computer_ids: Iterable[int],
) -> Dict[int, dict]:
    """
    Retrieve a mapping of {computer_id: extra_info_dict} for a list of computer ids
    based on parameters in validated_data.
    """

    extra_data = defaultdict(dict)

    if validated_data.get("with_network"):
        info = AccountNetworkInfo(account)
        computer_devices = info.get_devices_for_computer_ids(computer_ids)

        for computer_id, devices in computer_devices.items():
            extra_data[computer_id]["network_devices"] = sorted(
                (device.get_state() for device in devices),
                key=lambda x: x["interface"],
            )

    if validated_data.get("with_hardware"):
        management = HardwareInfoManagement(account)
        hardware = management.get_hardware_for_computer_ids(computer_ids)

        for computer_id, data in hardware.items():
            extra_data[computer_id]["hardware"] = data

    if validated_data.get("with_annotations"):
        management = AnnotationManagement(account, person)
        annotations = management.get_annotation(computer_ids)

        for computer_id, data in annotations.items():
            extra_data[computer_id]["annotations"] = data

    if validated_data.get("with_grouped_hardware"):
        management = HardwareInfoManagement(account)
        hardware = management.get_hardware_for_computer_ids(computer_ids)

        for computer_id, data in hardware.items():
            parser = HardwareInfoParser(computer_id, data)
            hardware = parser.hardware
            extra_data[computer_id]["grouped_hardware"] = hardware

    return extra_data


@with_computer(ComputersGetSchema, "ViewComputer")
async def computers_get_handler(
    request: Request,
    principal: Principal,
    computer: Computer,
    selection: ComputerSelection,
    validated_data: dict,
    url_args: typing.Mapping[str, int],
) -> ApiResponse:
    """Get information for a single computer"""

    computer_data = {computer.id: computer.serialize()}

    extra_data = _get_extra_computer_info(
        validated_data, principal.account, principal.person, [computer.id]
    )

    for computer.id, extra_data_for_computer in extra_data.items():
        computer_data[computer.id].update(extra_data_for_computer)

    return computer_data[computer.id], 200


@with_computer(ComputersPutSchema, "ViewComputer")
async def computers_put_handler(
    request: Request,
    principal: Principal,
    computer: Computer,
    selection: ComputerSelection,
    validated_data: typing.Mapping,
    url_args: typing.Mapping[str, int],
) -> ApiResponse:
    """Updates a single computer."""
    try:
        tag_names = get_tag_names(validated_data.get("tags", []))
    except InvalidTagModelError as e:
        raise ApiRequestError(str(e))

    access_context = None
    if validated_data["access_group"]:
        access_group = validated_data["access_group"]

        try:
            [access_context] = principal.account.get_access_contexts(
                names=[access_group]
            )
        except UnknownAccessContextsError:
            raise ApiRequestError(f"Access group '{access_group}' does not exist.")

    if validated_data["title"]:
        computer.title = validated_data["title"]

    if access_context:
        computer.access_context = access_context

    if tag_names:
        tags = set(get_tags(tag_names))
    else:
        tags = set()

    # Extra check for `None` as tags can be an empty list to clear them.
    if validated_data["tags"] is not None:
        computer.update_tags(set(tags))

    if validated_data["comment"] is not None:
        computer.comment = validated_data["comment"]

    return computer.serialize(), 200


ComputersActivitiesSchema = search_schema.extend(
    name="ComputersActivities",
    parameters=[
        Array(
            "computer_ids",
            Integer(),
            doc="Only results for the computers with these ids will be returned",
        ),
    ],
    errors=[],
    doc="""Details of the activities for the selected computers.""",
)


@login_required
@with_schema(ComputersActivitiesSchema)
@paginated
async def computers_activities_list_handler(
    request,
    principal,
    validated_data,
    limit=None,
    offset=None,
):
    account = get_account_by_id(principal.account_id)
    person = get_person_by_id(principal.id)
    query = validated_data.get("query")
    computer_ids = validated_data.get("computer_ids")
    access_contexts = get_access_context_ids_by_entity_type(
        account.access_context, person, "view"
    )
    account_store = account.store

    collection = ActivityChildCollection(
        account_store,
        account=account,
        entity_class=AccountComputerTwin,
        entity_ids=computer_ids,
    )
    twins = collection.find()

    activity_ids = twins.values(Activity.id)

    try:
        result = get_children_for_query(account, access_contexts, query, Activity.id)
    except InvalidSearchError as e:
        return {"error": "InvalidQueryError", "message": str(e)}, 400

    result.config(distinct=True)
    result = result.find(Activity.id.is_in(activity_ids))
    count = result.count()
    result.order_by(Desc(Activity.id))
    result.config(limit=limit, offset=offset)
    ids = list(result)
    children = (
        ActivityChildCollection(account=account, ids=ids)
        .find()
        .order_by(Desc(Activity.id))
    )

    results = [child.get_state() for child in children]
    return {"count": count, "results": results}, 200


@method
class AddAnnotationToComputers(AccountMethod):
    """
    Add a custom annotation to a selection of computers.

    This method takes two mandatory arguments:

    - **query**: A query string used to select the computers for which to add
      annotations.
      See :ref:`Computer Queries <computer-query>` for details.

    - **key**: Annotation key to add to the selected computers.
    - **value**: Annotation value associated with the provided key to add
      to the selected computers. (optional)

    For example, the following request will add a custom annotation
    representing physical to all computers running Ubuntu 10.04 (Lucid)::

        ?action=AddAnnotationToComputers&query=distribution:10.04&key=location
            &value=BLDG3:FLR2:CAGE101A
    """

    requires = "ViewAccount"

    route = SelectionRoute(permission="EditComputerDetails", error_if_empty=False)

    schema = Schema(
        name="AddAnnotationToComputers",
        parameters=[
            Unicode(
                "query",
                doc=(
                    "A query string used to select the computers to "
                    "which to add the annotation."
                ),
            ),
            Unicode("key", doc=("Annotation key to add to the selected computers.")),
            Unicode(
                "value",
                optional=True,
                doc=(
                    "Annotation value associated with the provided key to "
                    "add to the selected computers."
                ),
            ),
        ],
        errors=[InvalidQueryError, InvalidAnnotationKeyError],
        doc="""
    Add annotation key and optional value to a selection of computers.""",
    )

    def query(self, context, call):
        """
        Find all computers matching query and add the specified
        annotation for these computers.
        """
        person = get_person_by_id(call.principal.id)
        management = AnnotationManagement(context.account, person)
        annotation = {call.args.key: call.args.value}
        management.add_annotation(
            [computer.id for computer in context.get_selected()], annotation
        )
        return {}


@method
class RemoveAnnotationFromComputers(AccountMethod):
    """
    Remove a custom annotation with the specified key from a selection of
    computers.

    This method takes two mandatory arguments:

    - **query**: A query string used to select the computers to remove
      annotations from.
      See :ref:`Computer queries <computer-query>` for details.

    - **key**: The annotation key to remove.

    For example, the following request will remove the annotation key
    "location" from all computers with tag *server*::

        ?action=RemoveAnnotationFromComputers&query=tag:server&key=location
    """

    requires = "ViewAccount"

    route = SelectionRoute(permission="EditComputerDetails", error_if_empty=False)

    schema = Schema(
        name="RemoveAnnotationFromComputers",
        parameters=[
            Unicode(
                "query",
                doc=(
                    "A query string used to select the computers from "
                    "which to remove annotation."
                ),
            ),
            Unicode("key", doc="Annotation key to disassociate."),
        ],
        errors=[InvalidQueryError, InvalidAnnotationKeyError],
        doc="""Remove annotation key from a selection of computers.""",
    )

    def query(self, context, call):
        """
        Find all computers matching query and remove the specified
        annotation key for these computers.
        """
        person = get_person_by_id(call.principal.id)
        management = AnnotationManagement(context.account, person)
        management.remove_annotation(
            [computer.id for computer in context.get_selected()], [call.args.key]
        )
        return {}


class ComputerTagHandler(AccountMethod):
    requires = "ViewAccount"

    route = SelectionRoute(permission="EditTags", error_if_empty=False)

    def query(self, context, call):
        """
        Find all computers matching query and add the specified tags to the
        computers.
        """
        person = get_person_by_id(call.principal.id)
        for tag in call.args.tags:
            if not is_valid_tag(tag):
                raise InvalidTagError("Tag '%s' is invalid." % tag)
        self.method.__func__(
            person, get_tags(call.args.tags), list(context.get_selected())
        )
        return {}


@method
class AddTagsToComputers(ComputerTagHandler):
    """
    Add tags to a selection of computers.

    This method takes two mandatory arguments:

    - **query**: A query string used to select the computers to add tags to.
      See :ref:`Computer Queries <computer-query>` for details.

    - **tags**: Tag name to be applied, this can have more than one Tag, by
      numbering the tags with tags.1, tags.2, tags.3 etc.

    For example, the following request will add tags "server" and "lucid" to
    all computers running Ubuntu 10.04 (Lucid)::

        ?action=AddTagsToComputers&query=distribution:10.04&tags.1=server
            &tags.2=lucid
    """

    schema = Schema(
        name="AddTagsToComputers",
        parameters=[
            Unicode(
                "query",
                doc=("A query string used to select the computers to add tags to."),
            ),
            List("tags", Unicode(), doc="Tag names to be applied."),
        ],
        errors=[InvalidQueryError, InvalidTagError],
        doc="""Add tags to a selection of computers.""",
    )

    method = add_tags_to_computers


@method
class RemoveTagsFromComputers(ComputerTagHandler):
    """
    Remove tags from a selection of computers.

    This method takes two mandatory arguments:

    - **query**: A query string used to select the computers to remove tags
      from. See :ref:`Computer queries <computer-query>` for details.

    - **tags**: Tag name to be remove, this can have more than one Tag, by
      numbering the tags with tags.1, tags.2, tags.3 etc.

    For example, the following request will remove tags "server" and "lucid"
    from all computers with tag *server*::

        ?action=RemoveTagsFromComputers&query=tag:server&tags.1=server
            &tags.2=lucid
    """

    schema = Schema(
        name="RemoveTagsFromComputers",
        parameters=[
            Unicode(
                "query",
                doc=(
                    "A query string used to select the computers to "
                    "remove tags from."
                ),
            ),
            List("tags", Unicode(), doc="Tag names to be removed."),
        ],
        errors=[InvalidTagError, InvalidQueryError],
        doc="""Remove tags from a selection of computers.""",
    )
    method = remove_tags_from_computers


@method
class ChangeComputersAccessGroup(AccountMethod):
    """
    Change the access group for a selection of computers.

    This method takes two mandatory arguments:

    - **query**: A query string used to select the computers to change access
      group for.  See :ref:`Computer queries <computer-query>` for details.
    - **access_group**: The name of the access group to assign selected
      computers to.

    This is an example of a valid request::

         action=ChangeComputersAccessGroup&query=tag:new-servers
            &access_group=server

    The method returns a JSON serialized list of computers in the selection
    which have successfully changed access group::

         [{"access_group": "server",
           "id": 12345,
           "title": "A Computer",
           "comment": "",
           "total_memory": None,
           "total_swap": None,
           "reboot_required_flag": False,
           "hostname": "a_comp.example.com",
           "last_ping_time": None,
           "last_exchange_time": "2011-06-3017:59Z",
           "tags": ["server"],
           "network_devices": [{"broadcast_address": "192.168.1.255",
                                "interface": "eth0",
                                "ip_address": "192.168.1.2",
                                "mac_address": "00:1e:c9:6c:b8:de",
                                "netmask": "255.255.255.0"}]}]

    The following errors may be raised:

    - **InvalidQueryError**: If query format is invalid.
    """

    schema = Schema(
        name="ChangeComputersAccessGroup",
        parameters=[
            Unicode(
                "query",
                doc=(
                    "A query string used to select the computers to "
                    "change access group for."
                ),
            ),
            Unicode(
                "access_group",
                doc=(
                    "The name of the access group to assign selected " "computers to."
                ),
            ),
        ],
        errors={InvalidQueryError} | ComputerAccessContextChange.errors(),
        doc="""Change the access group for a selection of computers.""",
    )

    # Rely on the security checks in the service operation
    requires = "ViewAccount"

    route = SelectionRoute(permission=requires, error_if_empty=False)

    def query(self, context, call):
        account = context.account
        collection = AccountComputerTwinCollection(
            store=account.store,
            account=account,
            ids=context.get_selected_computer_ids(),
        )

        principal = PersonPrincipal(call.principal.id)
        operation = ComputerAccessContextChange(collection, principal)
        # Grab get_selected before we update_access_context. Once the access
        # group change occurs, a subsequent get_selected call will filter by
        # RemoveComputerFromAccessGroup permission. If the person doesn't also
        # have RemoveCompFromAG on the new group, our selection will disappear.
        selected_computers = list(context.get_selected())
        computer_ids = operation(call.args.access_group)
        return [
            computer.get_state()
            for computer in selected_computers
            if computer.id in computer_ids
        ]


@method
class RemoveComputers(AccountMethod):
    """
    Remove a list of computers by ID.

    This method takes a mandatory argument:

    - **computer_ids**: A list of computer ids to remove.

    This is an example of a valid request::

        ?action=RemoveComputers&computer_ids.1=30&computer_ids.2=43
    """

    schema = Schema(
        name="RemoveComputers",
        parameters=[
            List("computer_ids", Integer(), doc="A list of computer ids to remove.")
        ],
        errors=(list(RemoveComputersOperation.errors()) + [UnknownComputerError]),
        doc="""Remove a list of computers by ID.""",
    )

    route = AccountRoute()

    def query(self, context, call):
        """
        Remove computers with the specified ids.
        """
        collection = AccountComputerTwinCollection(context.store, account=context)
        collection = collection.with_ids(call.args.computer_ids)
        try:
            RemoveComputersOperation(collection, call.principal)()
        except UnknownEntityIDsError as error:
            raise UnknownComputerError("Unknown computer '{}'".format(error.ids.pop(0)))
        return {}


@method
class GetPendingComputers(AccountMethod):
    """
    Get a list of pending computers associated with the account used for
    authentication::

        ?action=GetPendingComputers

    The method returns a JSON serialized list of pending computers::

         [{"id": 12345,
           "title": "My Server",
           "hostname": "server.london.company.com",
           "creation_time": "2011-06-3017:59Z",
           "vm_info": "xen",
           "client_tags": ["london"], ["server"]}]
    """

    schema = Schema(
        name="GetPendingComputers",
        parameters=[],
        doc="""
    Get a list of pending computers associated with the account used for
    authentication.
    """,
    )

    requires = "ViewPendingComputer"

    route = AccountRoute()

    def query(self, context, call):
        """
        Find all L{PendingComputers} associated with an account.
        """
        account = context  # For readability
        person = get_person_by_id(call.principal.id)
        pending_computers = account.get_pending_computers(person)
        return [computer.get_state() for computer in pending_computers]


@method
class AcceptPendingComputers(AccountMethod):
    """
    Accept a list of pending computers associated with the account used for
    authentication.

    This method takes the following arguments:

    - **computer_ids**: A list of pending computer ids to accept.
    - **access_group**: The name of the access group to accept the computers
      to. If not provided, they will be put into the global access
      group. (optional)

    This is an example of a valid request::

        ?action=AcceptPendingComputers?computer_ids.1=1&computer_ids.2=2

    The method returns a JSON serialized list of accepted computers::

         [{"id": 12345,
           "title": "A Computer",
           "comment": "",
           "hostname": "a_comp.example.com",
           "last_exchange_time": "2011-06-3017:59Z"}]

    To replace existing computers, map the pending IDs to the existing
    ones::

        ?action=AcceptPendingComputers?computer_ids.1=1&computer_ids.2=2
            &existing_ids.2=3

    The following errors may be raised:

    - **InsufficientLicenses**: Insufficient licenses available to accept new
      computers.
    - **UnknownAccessGroup**: The access group is not known or the person
      is not authorized to accept pending computers into the access group.
    """

    schema = Schema(
        name="AcceptPendingComputers",
        parameters=[
            List("computer_ids", Integer(), doc="A list of computer IDs to accept."),
            Mapping(
                "existing_ids",
                Integer("pending", doc="The ID of the pending computer."),
                Integer("existing", doc="The ID of the computer to replace."),
                optional=True,
                doc=("A mapping from pending computer IDs " "to existing ones."),
            ),
            Unicode(
                "access_group",
                optional=True,
                default=None,
                doc="The access group to put the computers into",
            ),
        ],
        errors=[
            InsufficientLicensesError,
            UnknownComputerError,
            UnknownAccessGroupError,
        ],
        doc="""
    Accept a list of pending computers associated with the account used for
    authentication.""",
    )

    # We rely on the security checks in PendingComputer.accept to keep out the
    # baddies. Oh dear!
    requires = "ViewAccount"

    route = AccountRoute()

    def query(self, context, call):
        """
        Accept L{PendingComputer}s associated with an account.
        """
        account = context  # For readability
        person = get_person_by_id(call.principal.id)
        access_group = call.args.access_group
        pending_ids = call.args.computer_ids
        existing_ids = dict(call.args.existing_ids or ())

        extra = set(existing_ids) - set(pending_ids)
        if extra:
            raise UnknownComputerError(
                ("extra pending computer IDs in existing_ids: {}").format(
                    ",".join(str(cid) for cid in sorted(extra))
                )
            )

        pending = self._get_pending_computers(
            account, person, access_group, pending_ids, existing_ids
        )
        _check_computer_ids([p.computer for p in pending], pending_ids)
        _check_computer_ids(
            [p.existing for p in pending],
            sorted(existing_ids.values()),
            msg="Invalid 'existing' computer IDs: {}",
        )

        try:
            computers = account.accept_computers(person, pending)
        except LicenseSeatError:
            raise InsufficientLicensesError("Insufficient licenses to accept computers")
        except AcceptanceUnauthorized as e:
            raise UnauthorisedError(str(e))

        return sorted(
            [computer.get_state() for computer in computers],
            key=lambda x: list(x.values()),
        )

    def _get_pending_computers(
        self, account, person, access_group, computer_ids, existing_ids
    ):
        get_pending_computer = lambda cid: account.get_pending_computer_by_id(
            person, cid
        )
        get_existing_computer = lambda cid: account.get_computer(cid)

        # Make a closure for get_access_context().
        if access_group:
            access_context = self.get_access_context(account, access_group)
            get_access_context = lambda _: access_context
        else:
            default_access_context = self.get_access_context(
                account, DEFAULT_ACCESS_CONTEXT_NAME
            )

            def get_access_context(computer):
                if computer.access_group:
                    return self.get_access_context(account, computer.access_group)
                else:
                    return default_access_context

        # Return the list of PendingComputerToAccept.
        pending = self._iter_pending_computers(
            computer_ids,
            existing_ids,
            get_pending_computer,
            get_existing_computer,
            get_access_context,
        )
        return list(pending)

    def _iter_pending_computers(
        self, computer_ids, existing_ids, get_pending, get_existing, get_access_context
    ):
        for computer_id in computer_ids:
            existing_id = existing_ids.get(computer_id)
            if existing_id is not None:
                existing = get_existing(existing_id)
            else:
                existing = None

            pending_id = computer_id
            pending = get_pending(pending_id)

            if pending is not None:
                access_context = get_access_context(pending)
            else:
                access_context = None

            yield PendingComputerToAccept(pending, existing, access_context)


@method
class RejectPendingComputers(AccountMethod):
    """
    Reject a list of pending computers associated with the account used for
    authentication::

        ?action=RejectPendingComputers?computer_ids.1=1&computer_ids.2=2
    """

    schema = Schema(
        name="RejectPendingComputers",
        parameters=[
            List("computer_ids", Integer(), doc="A list of computer IDs to reject.")
        ],
        errors=[UnknownComputerError],
        doc="""
    Reject a list of pending computers associated with the account used for
    authentication.
    """,
    )

    requires = "RejectPendingComputer"

    route = AccountRoute()

    def query(self, context, call):
        """
        Reject L{PendingComputer}s associated with an account.
        """
        account = context  # For readability
        person = get_person_by_id(call.principal.id)

        computer_ids = call.args.computer_ids
        pending_computers = [
            account.get_pending_computer_by_id(person, computer_id)
            for computer_id in computer_ids
        ]
        _check_computer_ids(pending_computers, computer_ids)

        context.reject_computers(person, pending_computers)
        return {}


def _check_computer_ids(computers, requested, msg=None):
    if msg is None:
        msg = "Invalid computer_ids supplied: {}"
    requested = set(cid for cid in requested if cid is not None)

    computer_ids = set(x.id for x in computers if x)
    missing_ids = requested.difference(computer_ids)
    if missing_ids:
        raise UnknownComputerError(msg.format(",".join(map(str, sorted(missing_ids)))))


@method
class CreateCloudOtps(AccountMethod):
    """
    Create one-time passwords used for registration of cloud instances::

        ?action=CreateCloudOtps?count=3

    The method returns a JSON serialized list of one-time passwords, one for
    each requested::

         ["otp1", "otp2", "otp3"]

    You can then use those OTPs in the client configuration, using cloud-init
    for example.
    """

    schema = Schema(
        name="CreateCloudOtps",
        parameters=[
            Integer(
                "count", doc="The number of OTPs to create", optional=True, default=1
            )
        ],
        doc="""
    Create one-time passwords used for registration of cloud instances.
    """,
    )
    requires = "CreateComputer"

    route = AccountRoute()

    def query(self, context, call):
        result = []
        for i in range(call.args.count):
            otp = make_secure_id()
            result.append(otp)
            context.store.add(CloudRegistrationOTP(otp, account=context))
        return result


@method
class RebootComputers(AccountMethod):
    """
    Reboot a list of computers.

    This method takes a mandatory argument:

    - **computer_ids**: A list of computer ids to reboot.

    It also supports an optional argument:

    - **deliver_after**: Reboot the computer after the specified time.  The
      time format is YYYY-MM-DDTHH:MM:SSZ.

    This is an example of a valid request::

        ?action=RebootComputers&computer_ids.1=30&computer_ids.2=43

    The method returns a JSON serialized activity::

        {'computer_id': None,
         'creation_time': '2012-11-19T18:11:51Z',
         'creator': {'email': 'john@example.com', 'id': 3,
                      'name': 'John Smith'},
         'id': 141,
         'parent_id': None,
         'summary': 'Restart computer',
         'type': 'ActivityGroup'}
    """

    schema = Schema(
        name="RebootComputers",
        parameters=[
            List("computer_ids", Integer(), doc="A list of computer ids to reboot."),
            Unicode(
                "deliver_after",
                optional=True,
                doc="A time in the future to deliver the script.",
            ),
        ],
        errors=[InvalidTimeError, UnknownComputerError],
        doc="""Reboot a list of computers.""",
    )

    requires = "RestartComputer"

    route = SelectionRoute(permission=requires)

    def query(self, context, call):
        """
        Reboot computers with the specified ids.
        """
        person = get_person_by_id(call.principal.id)
        management = ComputerManagement(context, person)

        deliver_after = None
        if call.args.deliver_after:
            try:
                deliver_after = parse_datetime(call.args.deliver_after)
            except ValueError:
                raise InvalidTimeError(
                    "Invalid format for deliver_after: '%s'" % call.args.deliver_after
                )

        activity = management.restart(deliver_after=deliver_after)
        return activity.get_state()


@method
class ShutdownComputers(AccountMethod):
    """
    Shut down a list of computers.

    This method takes a mandatory argument:

    - **computer_ids**: A list of computer ids to shut down.

    It also supports an optional argument:

    - **deliver_after**: Shutdown the computer after the specified time.  The
      time format is YYYY-MM-DDTHH:MM:SSZ.

    This is an example of a valid request::

        ?action=ShutdownComputers&computer_ids.1=30&computer_ids.2=43

    The method returns a JSON serialized activity::

        {'computer_id': None,
         'creation_time': '2012-11-19T18:14:19Z',
         'creator': {'email': 'john@example.com', 'id': 3,
                      'name': 'John Smith'},
         'id': 147,
         'parent_id': None,
         'summary': 'Shutdown computer',
         'type': 'ActivityGroup'}
    """

    schema = Schema(
        name="ShutdownComputers",
        parameters=[
            List("computer_ids", Integer(), doc="A list of computer ids to shutdown."),
            Unicode(
                "deliver_after",
                optional=True,
                doc="A time in the future to deliver the script.",
            ),
        ],
        errors=[InvalidTimeError, UnknownComputerError],
        doc="""Shutdown a list of computers.""",
    )

    requires = "ShutdownComputer"

    route = SelectionRoute(permission=requires)

    def query(self, context, call):
        """
        Shutdown computers with the specified ids.
        """
        person = get_person_by_id(call.principal.id)
        management = ComputerManagement(context, person)

        deliver_after = None
        if call.args.deliver_after:
            try:
                deliver_after = parse_datetime(call.args.deliver_after)
            except ValueError:
                raise InvalidTimeError(
                    "Invalid format for deliver_after: '%s'" % call.args.deliver_after
                )

        activity = management.shutdown(deliver_after=deliver_after)
        return activity.get_state()


@method
class RenameComputers(AccountMethod):
    """
    Rename a set of computers.

    This method takes a single mandatory argument:

    - **computer_titles**: A mapping of computer_ids to titles to set.

    This is an example of a valid request::

        ?action=RenameComputers&computer_titles.30:newname
    """

    schema = Schema(
        name="RenameComputers",
        parameters=[
            Mapping(
                "computer_titles",
                Integer("computer_id", doc="The ID of the computer to rename"),
                Unicode("computer_name", doc="The new name to apply."),
                doc="mapping of computer_ids to computer titles",
            )
        ],
        errors=[UnknownComputerError, InvalidTitleError],
        doc="Rename a set of computers.",
    )

    requires = "EditComputerDetails"

    route = SelectionRoute(permission=requires)

    def query(self, context, call):
        """Rename a set of computers."""
        account = context
        computer_titles = dict(call.args.computer_titles)
        computers = account.get_computers_by_ids(list(computer_titles.keys()))
        for computer in computers:
            title = computer_titles[computer.id]
            # title is required
            if not title:
                raise InvalidTitleError("Invalid title: '{}'".format(title))
            computer.title = title

        return {}


@method
class GetWSLHosts(AccountMethod):
    """
    Gets a list of Windows computers that are hosting at least one WSL instance.
    """

    requires = "ViewAccount"

    schema = search_schema.extend(
        name="GetWSLHosts",
        parameters=[],
        errors=[],
        doc="Gets a list of Windows computers that are hosting at least one WSL instance",
    )

    route = SelectionRoute(permission="ViewComputer", error_if_empty=False)

    def query(self, context: ComputerSelection, call: Call):
        store = context._main_store
        origin = [
            Computer,
            Join(Distribution, Computer.distribution_id == Distribution.id),
            Join(ComputerRelationship, ComputerRelationship.parent_id == Computer.id),
        ]
        search = ComputerSearch(context.account)
        where = search.get_where_clause(call.args.query)

        result = store.using(*origin).find(
            Computer,
            And(*where),
            Distribution.code_name == WSL_HOST_CODE_NAME,
            ComputerRelationship.parent_id != None,
        )
        result = result.config(
            distinct=True,
            offset=call.args.offset,
            limit=call.args.limit,
        )

        return [computer.get_state() for computer in result]


WSLHostsListSchema = GetWSLHosts.schema.extend(name="WSLHostsList")


@login_required
@with_schema(WSLHostsListSchema)
@paginated
async def wsl_hosts_list_handler(
    request,
    principal,
    validated_data,
    limit=None,
    offset=None,
):
    permission = "landscape.ViewComputer"

    collection = ComputerCollection(
        account_id=principal.account_id,
        person_id=principal.id,
        permission=permission,
    )

    if "query" in validated_data and validated_data["query"]:
        try:
            collection = collection.with_search_criteria(validated_data["query"])
        except SearchCriteriaError as error:
            return {"error": "InvalidQueryError", "message": str(error)}, 400

    store = get_main_store()
    origin = [
        Computer,
        Join(Distribution, Computer.distribution_id == Distribution.id),
        Join(ComputerRelationship, ComputerRelationship.parent_id == Computer.id),
    ]
    computer_ids = store.using(*origin).find(
        Computer.id,
        Distribution.code_name == WSL_HOST_CODE_NAME,
        ComputerRelationship.parent_id != None,
    )

    result = collection.find()
    result = result.find(Computer.id.is_in(computer_ids))
    count = result.count()

    result.order_by(Computer.title)
    result.config(offset=offset, limit=limit)

    return {
        "count": count,
        "results": [computer.serialize() for computer in result],
    }, 200


@method
class RemoveWSLHosts(AccountMethod):
    """Removes Windows computers that are hosting at least one WSL instance."""

    schema = Schema(
        name="RemoveWSLHosts",
        parameters=[
            List("computer_ids", Integer(), doc="A list of computer ids to remove."),
            Bool(
                "cascade_to_children",
                optional=True,
                default=False,
                doc="If true, removed computers will also include child computers.",
            ),
        ],
        errors=[
            UnknownComputerError,
            NotWSLHostError,
            *RemoveComputersOperation.errors(),
        ],
        doc="Remove a list of Windows computers that are hosting at least one WSL instance by id",
    )

    route = AccountRoute()

    def query(self, context: Account, call: Call):
        """Remove WSL host computers with the specified ids."""

        # Determine if computers are WSL hosts before removing
        main_store = get_main_store()
        hosts = main_store.using(
            Computer,
            Join(Distribution, Computer.distribution_id == Distribution.id),
            Join(ComputerRelationship, ComputerRelationship.parent_id == Computer.id),
        ).find(
            Computer.id,
            Distribution.code_name == WSL_HOST_CODE_NAME,
            ComputerRelationship.parent_id.is_in(call.args.computer_ids),
        )
        computers = list(hosts)
        not_hosts = set(call.args.computer_ids) - set(computers)

        if not_hosts:
            raise NotWSLHostError(
                f"Computers with id={not_hosts} are not WSL hosts or unknown"
            )

        if call.args.cascade_to_children:
            computers += list(
                context.get_child_computers(computers).values(Computer.id)
            )

        collection = AccountComputerTwinCollection(context.store, account=context)
        collection = collection.with_ids(computers)

        try:
            RemoveComputersOperation(collection, call.principal)()
        except UnknownEntityIDsError as error:
            raise UnknownComputerError(f"Unknown computer '{error.ids.pop(0)}'")

        return {"message": f"Successfully removed computers with id={computers}"}


class ChildComputersOnlyMixin:
    """
    Mixin to provide a check that every computer id in the API call is indeed
    a child computer. Classes with this mixin should add `NotChildComputerError`
    to their schema's list of errors and call the `ensure_children_only` method
    to raise that error as appropriate.
    """

    def ensure_children_only(self, context, person):
        management = ComputerManagement(context, person)
        computer_ids = context.get_selected_computer_ids()
        non_child_computers = management.get_non_child_computers(computer_ids)
        if non_child_computers:
            raise NotChildComputerError(
                f"Computers with id={sorted(list(non_child_computers))} are not child computers "
                "or are unknown"
            )


class CloudInitConversionMixin:
    """
    A mixin to extract the cloud init file that has to be valid text.
    """

    def decode_cloud_init(self, cloud_init):
        try:
            cloud_init = cloud_init.decode("utf-8")
        except UnicodeDecodeError:
            raise CloudInitEncodingError()
        return cloud_init


@method
class CreateChildComputer(AccountMethod, CloudInitConversionMixin):
    """
    Send a message to a hostagent to create a child computer instance.
    """

    requires = "CreateComputer"

    schema = Schema(
        name="CreateChildComputer",
        parameters=[
            Integer(
                "parent_id",
                doc="The id of the parent computer.",
            ),
            Unicode(
                "computer_name",
                doc="The name of child computer to create.",
            ),
            Data(
                "cloud_init",
                optional=True,
                doc="b64 encoded cloud init file contents.",
            ),
            Unicode(
                "data_id",
                optional=True,
                doc="Name of the vault secret containing the cloud init file.",
            ),
            Unicode(
                "token",
                optional=True,
                doc="The authentication token to be passed to the secrets manager.",
            ),
        ],
        errors=[
            InvalidParameterCombinationError,
            NotWSLHostError,
            SecretNotFoundError,
            SecretsAuthenticationError,
            SecretsSealedError,
            UnknownComputerError,
        ],
        doc="""
        Create child computer instances on a parent host machine.
        If provided, the new instance will be created according to the cloud-init
        file specified, or the one stored with the specified vault secret.

        Specifying both a cloud init file and a vault secret will result in an error.
        """,
    )

    route = SelectionRoute(permission="CreateComputer", error_if_empty=False)

    def query(self, context, call):
        computer_name = call.args.computer_name
        cloud_init = call.args.cloud_init
        data_id = call.args.data_id
        token = call.args.token

        if cloud_init and data_id:
            raise InvalidParameterCombinationError(
                "The request cannot contain both the cloud_init and "
                "the data_id parameters."
            )

        account = context.account
        parent_id = call.args.parent_id
        computer_parent = account.get_computers(computer_ids=[parent_id]).one()
        if computer_parent is None:
            raise UnknownComputerError(f"Invalid parent_id supplied: {parent_id}")
        if computer_parent.distribution.code_name != WSL_HOST_CODE_NAME:
            raise NotWSLHostError(
                f"Computer with id={parent_id} is not a host computer or is unknown."
            )
        if data_id and not computer_parent.secrets_name:
            raise SecretNotFoundError(
                f"Computer with id={parent_id} has no secrets configured."
            )

        if cloud_init:
            cloud_init = self.decode_cloud_init(cloud_init)

        person = get_person_by_id(call.principal.id)
        handler = self.get_action_handler(
            "create-child-computer", context=context, person=person
        )
        state = handler.handle(
            computer_parent, computer_name, cloud_init, data_id, token
        )

        return state


@method
class DeleteChildComputers(AccountMethod, ChildComputersOnlyMixin):
    """
    Send a message to a hostagent to uninstall a child computer instance.
    """

    requires = "RemoveComputer"

    schema = Schema(
        name="DeleteChildComputers",
        parameters=[
            List(
                "computer_ids", Integer(), doc="A list of child computer ids to delete."
            )
        ],
        errors=[NotChildComputerError, UnknownComputerError],
        doc="""Delete a list of child computers by ID.""",
    )

    route = SelectionRoute(permission="RemoveComputer", error_if_empty=False)

    def query(self, context, call):
        person = get_person_by_id(call.principal.id)
        self.ensure_children_only(context, person)
        handler = self.get_action_handler(
            "delete-child-computers", context=context, person=person
        )
        state = handler.handle(computer_ids=context.get_selected_computer_ids())

        return state


@method
class SetDefaultChildComputer(AccountMethod):
    """
    Send a message to a hostagent to set a default child computer instance.
    """

    requires = "EditComputerDetails"

    schema = Schema(
        name="SetDefaultChildComputer",
        parameters=[
            Integer(
                "parent_id",
                doc="The id of the parent host computer.",
            ),
            Integer(
                "child_id",
                doc="The id of the child computer to set as default.",
            ),
        ],
        errors=[
            NotWSLHostError,
            UnknownComputerError,
            UnknownRelationshipError,
        ],
        doc="""
        Set the child computer instance with the provided id as the default one
        for the host parent.
        """,
    )

    route = SelectionRoute(permission="EditComputerDetails", error_if_empty=False)

    def query(self, context, call):
        account = context.account
        parent_id = call.args.parent_id
        computer_parent = account.get_computers(computer_ids=[parent_id]).one()
        if computer_parent is None:
            raise UnknownComputerError(f"Invalid parent_id supplied: {parent_id}")
        if computer_parent.distribution.code_name != WSL_HOST_CODE_NAME:
            raise NotWSLHostError(
                f"Computer with id={parent_id} is not a host computer or is unknown"
            )

        child_id = call.args.child_id
        computer_child = account.get_computers(computer_ids=[child_id]).one()
        if computer_child is None:
            raise UnknownComputerError(f"Invalid child_id supplied: {child_id}")

        parent_of_computer_child = computer_child.get_parent_computer()
        if not parent_of_computer_child:
            raise UnknownRelationshipError(f"Computer with id={child_id} has no parent")
        if parent_of_computer_child.id != parent_id:
            raise UnknownRelationshipError(
                f"Computer with id={parent_id} has no relationship to computer with id={child_id}"
            )

        person = get_person_by_id(call.principal.id)
        handler = self.get_action_handler(
            "set-default-child-computer", context=context, person=person
        )
        state = handler.handle(computer_child)

        return state


@method
class ShutdownHostComputer(AccountMethod):
    """
    Send a message to a hostagent to shutdown a host computer.
    """

    requires = "ShutdownComputer"

    schema = Schema(
        name="ShutdownHostComputer",
        parameters=[
            Integer(
                "parent_id",
                doc="The id of the parent host computer.",
            ),
        ],
        errors=[
            NotWSLHostError,
            UnknownComputerError,
        ],
        doc="""
        Send a message to shutdown a host computer.
        """,
    )

    route = SelectionRoute(permission="ShutdownComputer", error_if_empty=False)

    def query(self, context, call):
        account = context.account
        parent_id = call.args.parent_id
        host_computer = account.get_computers(computer_ids=[parent_id]).one()
        if host_computer is None:
            raise UnknownComputerError(f"Invalid parent_id supplied: {parent_id}")
        if host_computer.distribution.code_name != WSL_HOST_CODE_NAME:
            raise NotWSLHostError(
                f"Computer with id={parent_id} is not a host computer or is unknown"
            )

        person = get_person_by_id(call.principal.id)
        handler = self.get_action_handler(
            "shutdown-host-computer", context=context, person=person
        )
        state = handler.handle(host_computer)

        return state


@method
class StartChildComputers(AccountMethod, ChildComputersOnlyMixin):
    """
    Send a message to a hostagent to start a child computer instance.
    """

    requires = "RestartComputer"

    schema = Schema(
        name="StartChildComputers",
        parameters=[
            List(
                "computer_ids", Integer(), doc="A list of child computer ids to start."
            )
        ],
        errors=[NotChildComputerError, UnknownComputerError],
        doc="""Start a list of child computers by ID.""",
    )

    route = SelectionRoute(permission="RestartComputer", error_if_empty=False)

    def query(self, context, call):
        person = get_person_by_id(call.principal.id)
        self.ensure_children_only(context, person)
        handler = self.get_action_handler(
            "start-child-computers", context=context, person=person
        )
        state = handler.handle(computer_ids=context.get_selected_computer_ids())

        return state


@method
class StopChildComputers(AccountMethod, ChildComputersOnlyMixin):
    """
    Send a message to a hostagent to stop a child computer instance.
    """

    requires = "ShutdownComputer"

    schema = Schema(
        name="StopChildComputers",
        parameters=[
            List("computer_ids", Integer(), doc="A list of child computer ids to stop.")
        ],
        errors=[NotChildComputerError, UnknownComputerError],
        doc="""Stop a list of child computers by ID.""",
    )

    route = SelectionRoute(permission="RestartComputer", error_if_empty=False)

    def query(self, context, call):
        person = get_person_by_id(call.principal.id)
        self.ensure_children_only(context, person)
        handler = self.get_action_handler(
            "stop-child-computers", context=context, person=person
        )
        state = handler.handle(computer_ids=context.get_selected_computer_ids())

        return state


@method
class GetComputerProcesses(AccountMethod):
    requires = "ViewComputer"
    route = AccountRoute()
    schema = Schema(
        name="GetComputerProcesses",
        parameters=[
            Integer("computer_id", doc="The numerical ID of the computer."),
            Integer("offset", optional=True, doc="The number of items to skip."),
            PositiveInteger(
                "limit", optional=True, doc="The number of items per page."
            ),
        ],
        errors=[UnknownComputerError],
        doc="Gets the active processes for the specified computer.",
    )

    def query(self, context: Account, call: Call) -> list:
        computer = context.get_computer(call.args.computer_id)

        if not computer:
            raise UnknownComputerError(
                f"Invalid computer ID supplied: {call.args.computer_id}"
            )

        result = computer.get_active_processes()
        processes = result.config(offset=call.args.offset, limit=call.args.limit)

        return [process.get_state() for process in processes]


COMPUTER_PROCESSES_LIST_SCHEMA = Schema(
    name="ComputerProcessesList",
    parameters=[
        Integer(
            "offset",
            optional=True,
            doc="The number of items to skip.",
            default=RESULT_OFFSET,
        ),
        PositiveInteger(
            "limit",
            optional=True,
            doc="The number of items per page.",
            default=RESULT_LIMIT,
        ),
    ],
    errors=[UnknownComputerError],
    doc="Gets the active processes for the computer.",
)


@login_required
@with_schema(COMPUTER_PROCESSES_LIST_SCHEMA)
@paginated
async def computer_processes_list_handler(
    request: Request,
    principal: Principal,
    validated_data: dict,
    url_args: dict,
    search_args: Optional[Sequence[str]] = None,
    limit: int = RESULT_LIMIT,
    offset: int = RESULT_OFFSET,
):
    permission = "landscape.ViewComputer"

    collection = ComputerCollection(
        account_id=principal.account_id,
        person_id=principal.id,
        permission=permission,
    )
    computer_id = url_args["computer_id"]
    result = collection.find()
    computer = result.find(Computer.id.is_in([computer_id])).one()

    if not computer:
        return {
            "error": "UnknownComputerError",
            "message": f"Invalid computer ID supplied: {computer_id}",
        }, 404

    result = computer.get_active_processes()

    if search_args:
        columns = [ActiveProcess.pid, ActiveProcess.name]
        result = add_search_query(result, columns, search_args)

    count = result.count()
    processes = result.config(
        offset=validated_data.get("offset", offset),
        limit=validated_data.get("limit", limit),
    )

    return {
        "count": count,
        "results": [process.serialize() for process in processes],
    }, 200


def create_send_signal_activity(
    account: Account,
    person_id: int,
    computer_id: int,
    pids: Sequence[int],
    signal: str,
):
    """
    Helper function for creating an activity group that sends a signal to the
    specified computer processes.
    """
    computer = account.get_computer(computer_id)

    if not computer:
        raise UnknownComputerError(f"Invalid computer ID supplied: {computer_id}")

    processes = computer.get_active_processes_by_pids(pids)

    if processes.is_empty():
        raise ComputerProcessNotFoundError(
            f"Computer processes {', '.join(map(str, pids))} could not be found"
        )

    person = get_person_by_id(person_id)
    parent = account.create_activity(
        ActivityGroup,
        access_context=computer.access_context,
        creator=person,
        summary=SIGNAL_PROCESSES_SUMMARY,
    )

    for process in processes:
        process.signal(signal, parent)

    parent.settle()

    return parent


@method
class TerminateComputerProcesses(AccountMethod):
    requires = "EditComputerDetails"
    route = AccountRoute()
    schema = Schema(
        name="TerminateComputerProcesses",
        parameters=[
            Integer("computer_id", doc="The numerical ID of the computer."),
            List(
                "pids",
                Integer(),
                doc="A comma separated list of PIDs of the processes to send an TERM signal",
            ),
        ],
        errors=[ComputerProcessNotFoundError, UnknownComputerError],
        doc="Creates an activity to send a TERM signal to the specified computer processes",
    )

    def query(self, context: Account, call: Call) -> dict:
        activity = create_send_signal_activity(
            context,
            call.principal.id,
            call.args.computer_id,
            call.args.pids,
            SIGNAL_TERM,
        )

        return activity.get_state()


@method
class KillComputerProcesses(AccountMethod):
    requires = "EditComputerDetails"
    route = AccountRoute()
    schema = Schema(
        name="KillComputerProcesses",
        parameters=[
            Integer("computer_id", doc="The numerical ID of the computer"),
            List(
                "pids",
                Integer(),
                doc="A comma separated list of PIDs of the processes to send a KILL signal",
            ),
        ],
        errors=[ComputerProcessNotFoundError, UnknownComputerError],
        doc="Creates an activity to send a KILL signal to the specified computer processes",
    )

    def query(self, context: Account, call: Call) -> dict:
        activity = create_send_signal_activity(
            context,
            call.principal.id,
            call.args.computer_id,
            call.args.pids,
            SIGNAL_KILL,
        )

        return activity.get_state()


SIGNAL_PROCESS_SCHEMA = Schema(
    name="SignalProcess",
    parameters=[
        Integer("computer_id", doc="The numerical ID of the computer."),
        Array(
            "pids",
            Integer(),
            doc="A comma separated list of PIDs of the processes to send signal",
        ),
    ],
    errors=[ComputerProcessNotFoundError, UnknownComputerError],
    doc="Creates an activity to send a signal to the specified computer processes",
)


@login_required
@with_schema(SIGNAL_PROCESS_SCHEMA)
async def signal_computer_proc_post_handler(
    request: Request,
    principal: Principal,
    validated_data: dict,
    url_args: dict,
):
    account = get_account_by_id(principal.account_id)

    signals = {"kill": SIGNAL_KILL, "terminate": SIGNAL_TERM}
    try:
        activity = create_send_signal_activity(
            account,
            principal.id,
            validated_data["computer_id"],
            validated_data["pids"],
            signals[url_args["signal"]],
        )
    except (ComputerProcessNotFoundError, UnknownComputerError) as e:
        return {"error": "Unprocessable Content", "message": str(e)}, 422
    except KeyError:
        return {
            "error": "Signal arg '{}' invalid".format(url_args["signal"]),
            "message": f"Signal argument can only be the following {list(signals.keys())}",
        }, 400

    return activity.get_state(), 200
