# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
Event Log
============

Methods for retrieving or exporting an account's EventLog
"""

from datetime import datetime, timedelta
from typing import Optional, Sequence

from twisted.web.server import Request

from canonical.txapi.server.schema import Integer, PositiveInteger, Schema
from canonical.txapi.server.method import method

from canonical.landscape.api import RESULT_OFFSET, RESULT_LIMIT
from canonical.landscape.api.helpers import (
    paginated,
    ApiResponse,
    with_management_class,
    with_schema,
    add_search_query,
)
from canonical.landscape.api.jwt_auth import login_required
from canonical.landscape.api.methods import AccountMethod
from canonical.landscape.api.principal import Principal
from canonical.landscape.api.routes import AccountRoute
from canonical.landscape.model.eventlog.entry import EventLogEntry
from canonical.landscape.model.eventlog.management import EventLog
from canonical.landscape.service.security import PersonPrincipal
from canonical.landscape.service.eventlog.get import (
    GetEventLogEntries,
    GetEventLogEntriesCount,
)


@method
class GetEventLog(AccountMethod):
    """Retrieve event log entries for the account, ordered by creation time.

    - **days**: The number of days prior to today from which to fetch log
      entries.
    - **limit**: The maximum number of results returned by the method. It
      defaults to 1000.
    - **offset**: The offset inside the list of results.

    For example, the following returns the last 7 days for event log entries
    in the account, limiting the result to 100 entries::

        ?action=GetEventLog&days=7&limit=100


    The method returns a JSON representation of the log entries returned::

        [{"creation_time": '2014-09-16T18:37:45Z',
          "entity_id": 2,
          "entity_name": "WebServer1",
          "entity_type": "Computer",
          "id": 1,
          "message": "Added tag 'apache' to computer 'WebServer1'",
          "person_id", 3,
          "person_name": "John Smith"},
          ...
        ]
    """

    schema = Schema(
        name="GetEventLog",
        parameters=[
            PositiveInteger(
                "days",
                optional=True,
                doc=(
                    "The number of days prior to today from which to "
                    "fetch log entries. It defaults to 30 days."
                ),
            ),
            PositiveInteger(
                "limit",
                default=RESULT_LIMIT,
                optional=True,
                doc=(
                    "The maximum number of results returned by the "
                    "method. It defaults to %d." % RESULT_LIMIT
                ),
            ),
            Integer(
                "offset",
                default=RESULT_OFFSET,
                optional=True,
                doc="The offset inside the list of results.",
            ),
        ],
        errors=GetEventLogEntries.errors(),
        doc="Retrieve event log for the account.",
    )

    route = AccountRoute()

    def query(self, context, call):
        """
        Return the account C{EventLog} over the requested number of C{days}.
        """
        operation = GetEventLogEntries(context, call.principal)
        result = operation(
            days=call.args.days, offset=call.args.offset, limit=call.args.limit
        )
        return [log.get_state() for log in result]


EVENT_LOG_LIST_HANDLER_SCHEMA = Schema(
    name="EventLogListHandler",
    parameters=[
        PositiveInteger(
            "days",
            optional=True,
            default=30,
            doc=(
                "The number of days prior to today from which to "
                "fetch log entries. It defaults to 30 days."
            ),
        ),
    ],
    errors=GetEventLogEntries.errors(),
    doc="Retrieve event log for the account.",
)


@with_schema(EVENT_LOG_LIST_HANDLER_SCHEMA)
@with_management_class(EventLog)
@paginated
async def event_log_list_handler(
    request: Request,
    principal: Principal,
    management: EventLog,
    validated_data: Optional[dict] = None,
    search_args: Sequence = (),
    offset: int = RESULT_OFFSET,
    limit: int = RESULT_LIMIT,
) -> ApiResponse:
    days = validated_data.get("days")
    start_time = datetime.now() - timedelta(days=days)

    result = management.get_entries_after(start_time)

    if search_args:
        result = add_search_query(
            result,
            [
                EventLogEntry.entity_name,
                EventLogEntry.person_name,
                EventLogEntry.message,
            ],
            search_args,
        )

    count = result.count()
    result = result.config(limit=limit, offset=offset)

    response = {
        "count": count,
        "results": [log.get_state() for log in result],
    }

    return response, 200
