# Copyright 2005-2022 Canonical Limited. All rights reserved.

import base64
import inspect
import json
import math
import urllib
import warnings
from functools import wraps
from typing import Any, Callable, Optional, Sequence, Tuple, Union

from twisted.web.server import Request

from zope.component import getUtility
from zope.security.management import getSecurityPolicy

from canonical.security.status import is_disabled

from canonical.txapi.server.schema import (
    AbstractParameter as Parameter,
    InvalidParameterValueError,
    Schema,
    MissingParameterError,
)
from canonical.txapi.server.exception import APIError

from storm.exceptions import NotOneError
from storm.expr import Cast, Eq, Like, Or, SQL
from storm.store import ResultSet

from canonical.security.interfaces import IRoleRegistry
from canonical.security.role import UnknownRoleError
from canonical.landscape.model.activity.collection import ActivityChildCollection
from canonical.landscape.model.configuration import get_api_url
from canonical.landscape.model.eventlog.management import EventLog
from canonical.landscape.model.main.person import (
    Person,
    get_person_by_identity,
    get_person_by_id,
)
from canonical.landscape.model.main.account import (
    Account,
    get_account_by_name,
    get_account_by_id,
)
from canonical.landscape.model.main.tag import get_tag
from canonical.landscape.model.main.errors import UnknownAccessContextsError
from canonical.landscape.model.selection.selection import ComputerSelection
from canonical.landscape.api.error import (
    UnknownAccountError,
    MultipleAccountsError,
    UnknownUserError,
    UnknownAccessGroupError,
    UnknownRoleError as APIUnknownRoleError,
    UnauthorisedError,
)
from canonical.landscape.api.jwt_auth import login_required
from canonical.landscape.api.principal import Principal
from canonical.landscape.api.participation import PersonParticipation
from canonical.landscape.api.types import ApiResponse


EMPTY_BODY = ""
"""Empty response body for messages without content, e.g. 204"""


class APIHelper:
    """
    Helper class with methods raising appropriate API errors.
    """

    def check_unknown_names(self, names, entities, entity_name):
        """Check if invalid names are found and raise an error if so."""
        if names is None:
            names = set([])

        known_names = [entity.name for entity in entities]
        invalid_names = set(names).difference(known_names)

        if not invalid_names:
            return entities

        def to_capital_case(word):
            if word.islower():
                return word.capitalize()
            else:
                return word

        code = "Unknown" + "".join(
            [to_capital_case(word) for word in entity_name.split()]
        )
        message = "Unknown %s name%s: %s" % (
            entity_name,
            "s" if len(invalid_names) > 1 else "",
            ", ".join(sorted(invalid_names)),
        )
        raise APIError(400, code=code, message=message)

    def get_account_by_name(self, name):
        """
        Find the L{Account} called C{name}.

        @param name: The name of the L{Account}.

        @raises UnknownAccountError: raised if the name doesn't match an
            L{Account} in the system.
        """
        account = get_account_by_name(name)
        if account is None:
            raise UnknownAccountError("Unknown account name '%s'." % name)
        return account

    def get_account_by_person(self, person):
        """
        Find the L{Account} associated with a L{Person}.

        @param person: The person to get the account of.
        @type person: L{canonical.landscape.model.main.person.Person}
        @raises MultipleAccountsError: raised if the person supplied is
            attached to more than one account.
        @raises UnknownAccountError: raised if the person supplied is not
            attached to any accounts.
        """
        try:
            account = person.accounts.one()
        except NotOneError:
            raise MultipleAccountsError("Multiple accounts matching the given person.")
        if account is None:
            raise UnknownAccountError("No account matching the given person.")
        return account

    def get_person_by_identity(self, user_id):
        """
        Find the L{Person} identified by user id.

        @raises UnknownUserError: raised if the user id is not known.
        """
        person = get_person_by_identity(user_id)
        if person is None:
            raise UnknownUserError("Unknown user id: %s" % user_id)
        return person

    def get_role(self, role):
        """Return the role specified by name."""
        registry = getUtility(IRoleRegistry)
        try:
            return registry.get(role)
        except UnknownRoleError:
            raise APIUnknownRoleError("The role '%s' does not exist." % role)

    def get_access_context(self, account, name):
        """Return the C{AccessContext} in the account with specified name.

        @raises UnknownAccessGroupError: if no L{AccessContext} is found.
        """
        try:
            [access_context] = account.get_access_contexts(names=[name])
        except UnknownAccessContextsError:
            raise UnknownAccessGroupError("Unknown access group name '%s'" % name)
        return access_context

    def get_management(self, management_class, call):
        """
        Return a management class for the person and account specified by the
        call L{Principal}.
        """
        account = get_account_by_id(call.principal.account_id)
        person = get_person_by_id(call.principal.id)
        return management_class(account, person)

    def get_tags_by_names(self, names):
        """Return the L{Tags}s with the given C{names}.

        @param names: The names of the tags to return.

        @raises APIError: If some names do not result in existent tags.
        """

        tags = (get_tag(name) for name in names)
        tags = [tag for tag in tags if tag is not None]
        return self.check_unknown_names(names, tags, "tag")

    def event_log_from_call(self, call):
        """Create an C{EventLog} instance from a call."""
        person = get_person_by_id(call.principal.id)
        account = get_account_by_id(call.principal.account_id)
        return EventLog(account, person)


class File(Parameter):
    """
    Parameter representing a file's name and contents. The format is
    C{filename + "$$" + base64content}.
    """

    kind = "file"

    def parse(self, value):
        """
        This is called by the server to unpack the value after receiving it
        from the wire.
        """
        if isinstance(value, bytes):
            value = value.decode()

        (filename, contents) = value.split("$$", 1)
        try:
            binary_contents = base64.b64decode(contents)
        except TypeError:
            # For txapi, any param parse errors need to raise a ValueError.
            # Our invalid Base64 encoding becomes a ValueError
            raise ValueError()
        return (filename, binary_contents)


class Data(Parameter):
    """
    A parameter representing some base64 encoded information.
    """

    kind = "data"

    def parse(self, value):
        """
        This is called by the server to decode the information after receiving
        it from the wire.
        """
        try:
            decoded = base64.b64decode(value)
        except TypeError:
            # For txapi, any param parse errors need to raise a ValueError.
            # Our invalid Base64 encoding becomes a ValueError
            raise ValueError()
        return decoded


def paginated(handler: Callable) -> Callable:
    """`RequestHandler` decorator that injects `offset` and `limit` parameters into the handler and
    adds the `"count"`, `"next"`, and `"previous"` keys to the response dict. The handler itself is
    responsible for using the pagination parameters to limit the body of their results and to have a
    `"results"` key in their response body.

    Example::

        @paginated
        def my_handler(request, offset=0, limit=25):
            count = db.count(Record)
            records = db.query(Record, offset=offset, limit=limit)
            return 200, {
                "count": count,
                "result": list(records),
            }

    """

    @wraps(handler)
    async def wrapper(request: Request, *args, **kwargs):
        offset = _parse_int_request_arg(request, "offset")
        limit = _parse_int_request_arg(request, "limit")

        if limit is not None:
            kwargs.update({"limit": limit})

            if offset is None:
                offset = 0

            kwargs.update({"offset": offset})

        search_args = _parse_list_request_arg(request, "search")
        handler_params = inspect.signature(handler).parameters

        if search_args and handler_params.get("search_args"):
            kwargs.update({"search_args": search_args})

        body, status = await call_handler(handler, request, *args, **kwargs)

        if not (200 <= status < 300):
            # We don't need to do anything if the handler was not successful
            return body, status

        if "results" not in body:
            warnings.warn(
                '@paginated decorator only works on response bodies that contain the key `"results"`'
            )
            return body, status

        if "count" not in body:
            warnings.warn(
                '@paginated-decorator request handlers should include the key `"count"`, with the total number of records of the listed resource.'
            )

        next_url = build_next_url(
            request, limit=limit, offset=offset, count=body.get("count")
        )
        previous_url = build_previous_url(
            request, limit=limit, offset=offset, count=body.get("count")
        )

        body.update(
            {
                "next": next_url,
                "previous": previous_url,
            }
        )

        return body, status

    return wrapper


def build_next_url(
    request: Request, *, limit: int, offset: int, count: Optional[int] = None
) -> Union[str, None]:
    """Creates a url with `limit` and `offset` as query parameters, for the next page of results.

    :returns: The next page's url, or `None` if it would be beyond the end of `count` items.
    """

    if limit is None or count == 0:
        return None

    if count is None:
        count = math.inf

    new_offset = max(0, offset + limit)

    if new_offset >= count:
        return None

    return _build_paginated_url(request, limit=limit, offset=new_offset)


def build_previous_url(
    request: Request, *, limit: int, offset: int, count: Optional[int] = None
) -> str:
    """Creates a url with `limit` and `offset` as query parameters, for the previous page of
    results.

    :returns: The previous page's url, or `None` if it would be a negative offset.
    """

    if limit is None or count == 0:
        return None

    if offset <= 0:
        return None

    if count is None:
        count = math.inf

    new_offset = min(count - limit, offset - limit)
    new_offset = max(0, new_offset)

    return _build_paginated_url(request, limit=limit, offset=new_offset)


def _build_paginated_url(request: Request, *, limit: int, offset: int) -> str:
    """Creates a url with the provided `limit` and `offset` query parameters, preserving existing
    ones.
    """
    path = request.path.decode()
    request_args = {k: v[-1] for k, v in request.args.items()}
    request_args.update({b"limit": limit, b"offset": offset})
    params = urllib.parse.urlencode(request_args)
    api_url = get_api_url()

    return urllib.parse.urljoin(api_url, path) + f"?{params}"


def _extract_request_arg(request: Request, arg: Union[str, bytes]) -> list:
    """Extracts the query parameter `arg` from `request` as a list."""
    if isinstance(arg, str):
        arg = arg.encode()

    args = request.args.get(arg, [])

    if not isinstance(args, list):
        args = [args]

    return args


def _parse_int_request_arg(
    request: Request, arg: Union[str, bytes]
) -> Union[int, None]:
    """Extracts the query parameter `arg` from `request` and cast it to an `int`. If this fails,
    returns `None`.
    """
    args = _extract_request_arg(request, arg)

    # We reverse the list because when a single query parameter is specified more than once,
    # e.g., ?x=1&x=2, we use the last instance in other places in the code
    # (see for example canonical.landscape.api.resource.ServiceAPI.get_call_arguments).
    # In this case we use the last instance parseable as an int.
    for raw_arg in reversed(args):
        try:
            return int(raw_arg)
        except ValueError:
            pass  # Keep searching

    return None


def _parse_list_request_arg(request: Request, arg: Union[str, bytes]) -> list:
    """
    Extracts the query parameter in the form `?arg=value1,value2` and casts it
    into a list.
    """
    args = _extract_request_arg(request, arg)
    list_str = []

    for raw_arg in args:
        try:
            str_arg = raw_arg.decode()
        except AttributeError:
            continue

        list_str += str_arg.split(",")

    return list_str


def check_account_permission(account: Account):
    if is_disabled(account):
        raise UnauthorisedError(
            {
                "error": "UnauthorizedError",
                "message": "Account disabled",
            }
        )


def check_person_permission(person: Person, obj: Any, permissions: Sequence[str]):
    policy = getSecurityPolicy()(PersonParticipation(person))
    missing_permissions = [
        not policy.checkPermission(f"landscape.{p}", obj) for p in permissions
    ]

    if any(missing_permissions):
        raise UnauthorisedError(
            {
                "error": "UnauthorizedError",
                "message": "You do not have permission to perform this action",
            }
        )


def with_computer_selection(
    schema: Schema, permission: str
) -> Callable[[Callable], Callable]:
    """RequestHandler decorator that instantiates a ComputerSelection object
    and passes it to the handler, ensuring that the caller has the required
    permissions to use the selection.

    This decorator includes the use of the `login_required` and `with_schema`
    decorators.

    :param schema: the schema used for validation.
    :type schema: Schema
    :param permission: the permission for the `ComputerSelection`.
    :type param: str

    :return Callable: The decorator that wraps the actual function.
    """

    def decorator(handler: Callable) -> Callable:
        @wraps(handler)
        @login_required
        @with_schema(schema)
        async def wrapper(request: Request, principal: Principal, *args, **kwargs):
            try:
                account = get_account_by_id(principal.account_id)
                check_account_permission(account)

                person = get_person_by_id(principal.id)
                selection = ComputerSelection(
                    account,
                    person,
                    "landscape." + permission,
                )
                check_person_permission(person, selection, [permission])
            except UnauthorisedError as e:
                message = "Unauthorised"

                if e.args:
                    message = e.args[0]

                return message, e.status

            if kwargs.get("validated_data"):
                # This implies that with_schema was used prior
                data = kwargs["validated_data"]
                computer_ids = data.get("computer_ids", [])
                selection.select_computers_by_id(computer_ids)

                unselected = set(computer_ids) - selection.get_selected_computer_ids()

                if unselected:
                    return {
                        "error": "Unprocessable Content",
                        "message": f"Invalid computer ids: {', '.join(map(str, unselected))}",
                    }, 422

            return await handler(request, principal, selection, *args, **kwargs)

        return wrapper

    return decorator


def with_computer(schema: Schema, permission: str) -> Callable[[Callable], Callable]:
    """RequestHandler decorator that identifies a Computer from a <computer_id> in the
    url and passes it to the handler, ensuring that the caller has the required
    permissions to use the Computer.

    This decorator includes the use of the `login_required` and `with_schema`
    decorators.

    :param schema: the schema used for validation.
    :type schema: Schema
    :param permission: the permission for the `Computer`.
    :type param: str

    :return Callable: The decorator that wraps the actual function.
    """

    def decorator(handler: Callable) -> Callable:
        @wraps(handler)
        @login_required
        @with_schema(schema)
        async def wrapper(request: Request, principal: Principal, *args, **kwargs):
            try:
                account = get_account_by_id(principal.account_id)
                check_account_permission(account)

                person = get_person_by_id(principal.id)
                selection = ComputerSelection(
                    account,
                    person,
                    "landscape." + permission,
                )
                check_person_permission(person, selection, [permission])
            except UnauthorisedError as e:
                message = "Unauthorised"

                if e.args:
                    message = e.args[0]

                return message, e.status

            computer_id = kwargs["url_args"]["computer_id"]
            selection.select_computers_by_id([computer_id])

            if not selection.get_selected_computer_ids():
                return {
                    "error": "UnknownComputerError",
                    "message": f"Unknown computer id '{computer_id}'",
                }, 404

            computer = selection.get_selected()[0]

            return await handler(
                request, principal, computer, selection, *args, **kwargs
            )

        return wrapper

    return decorator


def with_management_class(
    management_class: type, permissions: Sequence[str] = []
) -> Callable[[Callable], Callable]:
    """RequestHandler decorator that instantiates a management object and passes it to the handler,
    ensuring that the caller has the required permissions to use the manager.
    This necessitates the application the `login_required` decorator, so it also applies that.

    :param type management_class: A subclass of `ModelManagement` to be instantiated and passed
        to the wrapped callable.

    :return Callable: The decorator that wraps the actual function.
    """

    def decorator(handler: Callable) -> Callable:
        @wraps(handler)
        @login_required
        async def wrapper(request: Request, principal: Principal, *args, **kwargs):
            try:
                account = get_account_by_id(principal.account_id)
                check_account_permission(account)

                person = get_person_by_id(principal.id)
                management = management_class(account, person)
                check_person_permission(person, management, permissions)
            except UnauthorisedError as e:
                message = "Unauthorised"

                if e.args:
                    message = e.args[0]

                return e.args[0], e.status

            return await call_handler(
                handler, request, principal, management, *args, **kwargs
            )

        return wrapper

    return decorator


def with_schema(schema: Schema) -> Callable[[Callable], Callable]:
    """RequestHandler decorator that passes the request content through validation with `schema`,
    returning errors instead of invoking the wrapped handler.

    :param Schema schema: The schema to use for validation.

    :return Callable: The decorator that wraps the actual function.
    """

    def decorator(handler: Callable) -> Callable:
        @wraps(handler)
        async def wrapper(request: Request, *args, **kwargs):
            data = {}

            method = request.method.decode()

            if method in ("GET", "DELETE"):
                try:
                    data = {
                        key.decode(): value[0].decode()
                        for key, value in request.args.items()
                    }
                except UnicodeDecodeError:
                    return {
                        "error": "UnicodeDecodeError",
                        "message": "Query parameters are not valid UTF-8",
                    }, 400
            elif method in ("POST", "PUT", "PATCH"):
                try:
                    decoded_content = request.content.read().decode()

                    if decoded_content:
                        data = json.loads(decoded_content)
                except UnicodeDecodeError:
                    return {
                        "error": "UnicodeDecodeError",
                        "message": "Request body is not valid UTF-8",
                    }, 400
                except json.decoder.JSONDecodeError:
                    return {
                        "error": "JSONDecodeError",
                        "message": "Request body is not valid JSON",
                    }, 400

            try:
                validated_data, _ = schema.extract(data)
            except MissingParameterError as error:
                return {"error": "MissingParameterError", "message": error.message}, 400
            except InvalidParameterValueError as error:
                return {
                    "error": "InvalidParameterValueError",
                    "message": error.message,
                }, 400

            request.content.seek(0)

            return await handler(
                request,
                *args,
                validated_data=dict(validated_data),
                **kwargs,
            )

        return wrapper

    return decorator


async def call_handler(handler: Callable, *args, **kwargs) -> ApiResponse:
    """A utility for `RequestHandler` decorators so they can wrap both async and non-async
    `RequestHandler`s.
    """
    if inspect.iscoroutinefunction(handler):
        return await handler(*args, **kwargs)

    return handler(*args, **kwargs)


def get_activities_by_ids(store, account, ids):
    collection = ActivityChildCollection(store, account=account, ids=ids)

    activities = collection.find()
    activities.config(distinct=True)

    return activities


class ApiRequestError(Exception):
    """
    Raising this in the api handlers will return a default error json response. It
    can be customized by either subclassing or passing in message and/or code
    """

    message = "Please check your request and try again"
    code = 400

    def __init__(self, message=None, code=None):
        if message:
            self.message = message
        if code:
            self.code = code


def add_search_query(
    resultset: ResultSet,
    columns: Sequence,
    search_args: Sequence[str],
) -> ResultSet:
    """
    Helper function that takes a `ResultSet` and applies a search query by
    adding a `LIKE` argument to the SQL query.
    """
    query = []

    for search_arg in search_args:
        for column in columns:
            query.append(
                Like(
                    Cast(column, "TEXT"),
                    f"%{search_arg.lower()}%",
                    case_sensitive=False,
                )
            )

    if query:
        resultset = resultset.find(Or(*query))

    return resultset
