import transaction

from twisted.web.server import Request
from canonical.lib.mail import is_email_address
from canonical.lib.timestamp import format_datetime
from canonical.lib.validators import is_valid_timezone_string
from canonical.txapi.server.schema import Schema, Unicode

from canonical.landscape.api.helpers import with_schema
from canonical.landscape.api.principal import Principal
from canonical.landscape.api.jwt_auth import login_required
from canonical.landscape.api.types import ApiResponse
from canonical.landscape.model.access.management import AccessControlManagement
from canonical.landscape.model.configuration import get_api_url
from canonical.landscape.model.main.account import get_account_by_name
from canonical.landscape.model.main.person import get_person_by_id


generate_api_credentials_schema = Schema(
    name="GenerateAPICredentials",
    parameters=[
        Unicode("account"),
    ],
)

update_person_schema = Schema(
    name="UpdatePerson",
    parameters=[
        Unicode("name", optional=True),
        Unicode("email", validator=is_email_address, optional=True),
        Unicode("timezone", validator=is_valid_timezone_string, optional=True),
        Unicode("preferred_account", optional=True),
    ],
)

INVALID_ACCOUNT_BODY = {
    "error": "InvalidAccountError",
    "message": "You are not a member of that account.",
}


@login_required
async def api_credentials_get_handler(
    request: Request, principal: Principal
) -> ApiResponse:
    person = get_person_by_id(principal.id)
    accounts = list(person.accounts)
    credentials = [serialize_credentials(person, account) for account in accounts]

    return {
        "credentials": credentials,
    }, 200


def serialize_credentials(person, account):
    credentials = person.get_api_credentials(account)
    endpoint = get_api_url()

    if not credentials:
        return {
            "access_key": "",
            "account_name": account.name,
            "account_title": account.title,
            "endpoint": endpoint,
            "exports": "",
            "secret_key": "",
        }

    access_key_id = credentials.access_key_id.decode()
    access_secret_key = credentials.access_secret_key.decode()
    exports = "\n".join(
        [
            f"export LANDSCAPE_API_URI='{endpoint}'",
            f"export LANDSCAPE_API_KEY='{access_key_id}'",
            f"export LANDSCAPE_API_SECRET='{access_secret_key}'",
        ]
    )

    return {
        "access_key": access_key_id,
        "account_name": account.name,
        "account_title": account.title,
        "endpoint": endpoint,
        "exports": exports,
        "secret_key": access_secret_key,
    }


@login_required
@with_schema(generate_api_credentials_schema)
async def api_credentials_post_handler(
    request: Request, principal: Principal, validated_data: dict
) -> ApiResponse:
    person = get_person_by_id(principal.id)
    account_name = validated_data.get("account")
    account = get_account_by_name(account_name)

    if not account or person not in account.people:
        return INVALID_ACCOUNT_BODY, 400

    if person.get_api_credentials(account) is not None:
        person.delete_api_credentials(account)
    person.create_api_credentials(account)

    return serialize_credentials(person, account), 200


@login_required
async def person_get_handler(request: Request, principal: Principal) -> ApiResponse:
    person = get_person_by_id(principal.id)
    body = person.serialize()
    body["accounts"] = accounts_details(person)

    return body, 200


def accounts_details(person):
    results = []

    for account in person.accounts:
        management = AccessControlManagement(account)
        roles = management.get_persons_roles().get(person, [])
        details = {
            "name": account.name,
            "title": account.title,
            "roles": sorted(role.name for role in roles),
        }
        results.append(details)

    return results


@login_required
@with_schema(update_person_schema)
async def person_post_handler(
    request: Request, principal: Principal, validated_data: dict
) -> ApiResponse:
    person = get_person_by_id(principal.id)
    name = validated_data.get("name", "")
    email = validated_data.get("email", "")
    timezone = validated_data.get("timezone", "")
    preferred_account_name = validated_data.get("preferred_account", "")

    if preferred_account_name:
        preferred_account = get_account_by_name(preferred_account_name)
        if preferred_account is None:
            return {
                "error": "InvalidPreferredAccountError",
                "message": "Invalid account.",
            }, 400
        person.preferred_account = preferred_account

    if email and not person.is_allowable_email(email):
        return {
            "error": "InvalidEmailError",
            "message": "Invalid email.",
        }, 400

    if name and name.strip():
        person.name = name.strip()

    if email and email.strip():
        person.email = email.strip()

    if timezone:
        person.timezone = timezone

    return person.serialize(), 200
