# Copyright 2005-2024 Canonical Limited. All rights reserved.

import json
import logging
import traceback
from base64 import b64encode
from typing import Awaitable, Union, Tuple

from storm.twisted.transact import transact, RETRIABLE_ERRORS
from twisted.internet.defer import ensureDeferred
from twisted.logger import Logger
from twisted.web.server import NOT_DONE_YET, Request
from werkzeug.routing import Map, Submount
from werkzeug.exceptions import NotFound, MethodNotAllowed

from canonical.txapi.server.exception import APIError
from canonical.txapi.server.resource import QueryAPI
from canonical.txapi.server.schema import Schema, Unicode, Integer, Date, RawStr

from canonical.landscape.api.error import UnauthorisedError, MultiError
from canonical.landscape.api.helpers import call_handler, EMPTY_BODY, ApiRequestError
from canonical.landscape.api.jwt_auth import JwtException, jwt_authenticate_request
from canonical.landscape.api.principal import Principal, AnonymousCredentials
from canonical.landscape.api.types import ApiResponse, RequestHandler
from canonical.landscape.api.urls import APIS, DEFAULT_API_VERSION

from canonical.landscape.model.configuration import get_root_origin, is_development_mode
from canonical.landscape.model.main.person import get_credentials_by_access_key_id


log = Logger()


def request_get_arg(request, arg):
    arg_list = request.args.get(arg.encode())
    arg = arg_list[-1].decode() if arg_list else None
    return arg


class ServiceAPI(QueryAPI):
    """An API used to configure the Landscape service.

    @param transactor: A L{Transactor} that can be used to execute
        functions that run in a database transaction.
    """

    signature_versions = (2,)
    content_type = "application/json"

    schema = Schema(
        parameters=[
            Unicode("action"),
            RawStr("access_key_id", optional=True),
            Date("timestamp", optional=True),
            Date("expires", optional=True),
            Unicode("version", optional=True),
            Unicode("signature_method", optional=True, default="HmacSHA256"),
            Unicode("signature", optional=True),
            Integer("signature_version", optional=True, default=2),
        ]
    )

    def __init__(
        self,
        transactor,
        registry=None,
        path="",
    ):
        super().__init__(registry=registry, path=path)

        self.transactor = transactor

        # If for example /v2/ is left out from url, we use the default api version
        version_level = [Submount("/", APIS[DEFAULT_API_VERSION])]

        # Assign each api version to the prefix e.g. "v2" -> /v2
        for api_version, rules in APIS.items():
            version_level.append(Submount("/" + api_version, rules))

        # There is an api root at self.path for example /api/ but if left out then just
        # the root / will get routed correctly
        api_root = self.path.strip("/")
        self.url_map = Map(
            [
                Submount("/", version_level),
                Submount("/" + api_root, version_level),
            ]
        )

    def get_call_arguments(self, request):
        """
        Overridden from the base L{QueryAPI.get_call_arguments} because we use
        a different naming scheme for our core API parameters.

        First, parameter names are lowercase and underscore-separated instead
        of CamelCase; second, we want to change the "AWSAccessKeyId" to a
        simple "access_key_id" to avoid Amazon-specific terminology in our API.
        """
        params = {k.decode(): v[-1].decode() for k, v in request.args.items()}
        args, rest = self.schema.extract(params)
        # Get rid of Signature so it doesn't mess with signature verification
        params.pop("signature", "")
        result = {
            "transport_args": {
                "action": args.action,
                "access_key_id": args.access_key_id,
                "timestamp": args.timestamp,
                "expires": args.expires,
                "version": args.version,
                "signature_method": args.signature_method,
                "signature": args.signature,
                "signature_version": args.signature_version,
            },
            "handler_args": rest,
            "raw_args": params,
        }
        return result

    def get_method(self, call):
        return QueryAPI.get_method(self, call, self.transactor)

    @transact
    def get_principal(self, access_key: bytes, request: Request) -> Principal:
        """Creates a `Principal` using the `request`, falling back to using the `access_key`, for
        backwards-compatibility reasons.

        :param bytes access_key: The key to use if JWT authentication fails.
        :param request Request: The incoming HTTP request.
        :returns Principal: An authenticated or anonymous principal, depending on authentication
            results.
        """
        principal = self._get_principal_jwt(request)

        if principal is not None:
            # We got a real user from the JWT - relax signature validation.
            self._extra_validators = []
            return principal

        self._extra_validators = [self._validate_signature]

        return self._get_principal_access_key(access_key or "")

    def authorize(self, method, call):
        if not method.authorize(call):
            raise UnauthorisedError("Permission denied.")

    def dump_result(self, result):
        return json.dumps(self._decode_bytes(result))

    def dump_error(self, error, request):
        """
        Convert an exception to a json structure.

        L{APIError}s and L{MultiError}s can be converted.
        """
        if isinstance(error, MultiError):
            reasons = [
                {"message": reason.message, "param": reason.param, "error": reason.code}
                for reason in error.get_reasons()
            ]
            dump = {"error": error.code, "errors": reasons, "message": error.message}
        else:
            dump = {"message": error.message}
            if error.code:
                dump["error"] = error.code

        return json.dumps(dump)

    def route_path(self, request):
        """
        It routes the url path through werkzeug.
        If it cannot find a route or the method is not allowed, a 4XX handler is returned

        Returns the function handler and a dict containing args present in the url (if any)
        """
        try:
            if isinstance(request.path, str):
                path = request.path
            else:
                path = request.path.decode()
            rule, args = self.url_map.bind("").match(path, request.method.decode())
        except NotFound:
            return _404_handler, {}
        except MethodNotAllowed:
            return _405_handler, {}

        return rule, args

    async def handle_path(self, handler, url_args, request):
        """
        Calls the handler with the request and the parsed url args (if any)
        """
        try:
            if url_args:
                body, status = await call_handler(handler, request, url_args=url_args)
            else:
                body, status = await call_handler(handler, request)
            if body is not EMPTY_BODY:
                json_body = json.dumps(body).encode()
            else:
                json_body = EMPTY_BODY
            _write_response(
                request, json_body, status
            )  # Note an exception can also occur here
        except ApiRequestError as exc:
            body = {
                "error": type(exc).__name__,
                "message": exc.message,
            }
            json_body = json.dumps(body).encode()
            _write_response(request, json_body, exc.code)
            raise
        except Exception as exc:
            logging.error(traceback.format_exc())
            traceback.print_exc()  # Show the error during tests (logs are disabled)
            status = 500
            json_body = json.dumps(
                {"error": "InternalServerError", "message": "Server error"}
            ).encode()
            _write_response(request, json_body, status)
            raise

    def is_legacy_api(self, request):
        """
        Checks if the request is using the legacy api by looking at action and if version
        contains two dashes like for example 2011-08-01
        """
        action = request_get_arg(request, "action")
        if action:
            version = request_get_arg(request, "version")
            return version and version.count("-") == 2
        return False

    async def render_method_helper(self, request: Request) -> None:
        """
        Splitting this out as an async for unit tests
        """
        handler, url_args = self.route_path(request)
        await self.handle_path(handler, url_args, request)

    def render(self, request):

        # Override for dev mode, where we set the appropriate headers
        request.setHeader("Access-Control-Allow-Origin", get_root_origin())
        request.setHeader("Access-Control-Allow-Headers", "Content-Type, Authorization")
        if is_development_mode():
            request.setHeader("Access-Control-Allow-Origin", "*")
            request.setHeader(
                "Access-Control-Allow-Methods", "GET, POST, PUT, PATCH, DELETE, OPTIONS"
            )

        return super().render(request)

    def render_method(self, request: Request) -> int:
        """Run the request in a transaction in a separate thread"""

        def run_method():
            """
            An analog to the logic in Transactor._wrap to retry the transaction on
            retriable errors and abort immediately otherwise.

            ensureDeferred drives the async render_method_helper, but an unhandled
            exception in render_method_helper (i.e., one that isn't passed to an errback)
            will introduce an 'unhandled error in Deferred' into the reactor.

            Exceptions in render_method_helper *need* to be raised to transactor.run
            to abort the transaction, so we catch them in an errback to satisfy the
            twisted reactor and then raise any failures to transactor.run to abort
            the transaction.
            """
            d = ensureDeferred(self.render_method_helper(request))
            d.addErrback(get_exception)
            if isinstance(d.result, Exception):
                raise d.result

        def get_exception(failure):
            """
            Return exception instances instead of raising them to avoid creating
            unhandled errors in the Deferred created in run_method::ensureDefered.
            """
            return failure.value

        deferred = self.transactor.run(run_method)

        # Exceptions have already been logged and handled, so we ignore anything
        # that's been re-raised from the transactor.
        deferred.addErrback(lambda failure: None)

        return NOT_DONE_YET

    def render_POST(self, request: Request) -> int:
        """All post requests are transactional by default."""

        # Legacy api is handled by the parent class
        if self.is_legacy_api(request):
            return super().render_POST(request)

        return self.render_method(request)

    def render_GET(self, request) -> int:

        if self.is_legacy_api(request):
            return super().render_GET(request)

        return self.render_method(request)

    def render_PATCH(self, request) -> int:
        return self.render_method(request)

    def render_PUT(self, request) -> int:
        return self.render_method(request)

    def render_DELETE(self, request) -> int:
        return self.render_method(request)

    def render_OPTIONS(self, request) -> bytes:
        return b""

    def render_HEAD(self, request):
        """
        Return an empty response for HEAD requests for health checking
        """
        request.setHeader("Content-Type", "text/plain")
        request.setResponseCode(200)
        return b""

    def _decode_bytes(self, data):
        """
        Recursively decodes bytes to strings to prep for json dumping.
        If UTF-8 decoding doesn't work, b64-encodes first.
        """
        if isinstance(data, bytes):
            logging.warn("API endpoint returns bytes: %s", data)
            try:
                return data.decode()
            except UnicodeDecodeError:
                return b64encode(data).decode()

        if isinstance(data, list):
            return [self._decode_bytes(d) for d in data]

        if isinstance(data, dict):
            return {
                self._decode_bytes(k): self._decode_bytes(v) for k, v in data.items()
            }

        return data

    @staticmethod
    def _get_principal_access_key(access_key: bytes) -> Union[Principal, None]:
        """Retrieves the `Principal` matching the `access_key`, unless there is
        none.
        """
        if access_key == b"anonymous":
            return Principal(AnonymousCredentials())

        credentials = get_credentials_by_access_key_id(access_key)

        if credentials is None:
            return None

        return Principal(credentials)

    @staticmethod
    def _get_principal_jwt(request: Request) -> Union[Principal, None]:
        """Extracts the JWT from `request` uses it to retrieve the `Principal`.

        :param Request request: The incoming HTTP request.
        :returns Principal or None: The principal if authentication succeeds. Otherwise, None.
        """
        auth_header = request.getHeader("Authorization")

        if isinstance(auth_header, bytes):
            auth_header = auth_header.decode()

        if not auth_header or not auth_header.startswith("Bearer "):
            return None

        try:
            return jwt_authenticate_request(request)
        except JwtException as exc:
            raise APIError(401, exc.code, exc.message)


def _404_handler(_: Request) -> ApiResponse:
    return {"error": "NotFoundError", "message": "Not found"}, 404


def _405_handler(_: Request) -> ApiResponse:
    return {"error": "MethodNotAllowedError", "message": "Method not allowed"}, 405


def _write_response(request: Request, json_body: bytes, status: int) -> None:
    """Writes `response` into `request`, appropriately setting response HTTP headers.

    :param Request request: The incoming HTTP request.
    :param ApiResponse response: The response returned from a `RequestHandler`.
    """

    if json_body is EMPTY_BODY:
        return _write_empty_body_response(request, status)

    request.setResponseCode(status)
    request.setHeader("Content-Length", str(len(json_body)))
    request.setHeader("Content-Type", ServiceAPI.content_type)

    request.write(json_body)
    request.finish()


def _write_empty_body_response(request: Request, status: int) -> None:
    """Set the response code and set HTTP headers for an empty body"""
    request.setResponseCode(status)
    request.finish()
