# Copyright 2005-2024 Canonical Limited. All rights reserved.

from canonical.txapi.server.schema import (
    AbstractParameter as Parameter,
    List,
    Unicode,
    UnknownParametersError,
)

from canonical.lib.text import is_valid_title


class UnicodeTitle(Unicode):
    """
    A parameter that must be a C{unicode} string, without newlines, and contain
    more than just whitespace.
    """

    kind = "unicode title"

    def parse(self, value):
        value = super(UnicodeTitle, self).parse(value)
        if not is_valid_title(value):
            raise ValueError()
        return value


class Mapping(Parameter):
    """A parameter that maps keys to values."""

    kind = "mapping"
    supports_multiple = True

    def __init__(
        self, name=None, key=None, value=None, optional=False, default=None, doc=None
    ):
        """
        @param key: A Parameter instance which will be used to parse and
            format the keys in the mapping (default: Unicode).
        @param value: A Parameter instance which will be used to parse and
            format the values in the mapping (default: Unicode).
        """
        if key is None:
            key = Unicode()
        elif not hasattr(key, "parse") or not hasattr(key, "format"):
            raise ValueError("expected a Parameter for key")
        if value is None:
            value = Unicode()
        elif not hasattr(value, "parse") or not hasattr(value, "format"):
            raise ValueError("expected a Parameter for value")
        if default is None:
            default = {}
        super(Mapping, self).__init__(name, optional=optional, default=default, doc=doc)
        if key.name is None:
            key.name = name
        self.key = key
        if value.name is None:
            value.name = name
        self.value = value

    def parse(self, arg):
        """Convert the keys and values of the given mapping."""
        try:
            getpairs = arg.items
        except AttributeError:
            items = arg
        else:
            items = getpairs()
        parse_key = self.key.parse
        parse_value = self.value.parse
        result = {}
        for key, value in items:
            key = parse_key(key)
            value = parse_value(value)
            result[key] = value
        # txapi.server.schema.Arguments() does funny stuff when given
        # dicts, so we return the item pairs instead.
        return list(result.items())

    def format(self, arg):
        """Format the keys and values of the given mapping."""
        try:
            getpairs = arg.items
        except AttributeError:
            items = arg
        else:
            items = getpairs()
        format_key = self.key.format
        format_value = self.value.format
        result = {}
        for key, value in items:
            key = format_key(key)
            value = format_value(value)
            result[key] = value
        return result

    def typify(self, include_name: bool = False) -> dict:
        """See `canonical.txapi.server.schema.AbstractParameter.typify`."""
        result = super().typify(include_name=include_name)
        result["key"] = self.key.typify()
        result["value"] = self.value.typify()

        return result


class Array(List):
    """
    A homogenous list of instances of a parameterized type.

    The array can be provided as a comma separated list (e.g. when sending
    the array as a query parameter in a GET request), or as a JSON array
    (e.g. when sending the array as part of the body of a POST request).
    """

    kind = "array"

    def parse(self, value):
        if isinstance(value, str):
            values = value.split(",")
        elif isinstance(value, list):
            values = value
        else:
            raise ValueError(value)

        return [self.parse_item(value) for value in values]

    def parse_item(self, value):
        try:
            return self.item.coerce(value)
        except UnknownParametersError as error:
            # A nested `Structure` error.
            raise ValueError(error)
