# Copyright 2024 Canonical Limited. All rights reserved.

import os
from typing import Sequence

from storm.locals import Store
from twisted.web.server import Request
from requests import get, HTTPError, ConnectionError, Response

from canonical.txapi.server.schema import (
    Bool,
    Date,
    Integer,
    InvalidParameterValueError,
    Schema,
    Structure,
    Unicode,
    Union,
)

from canonical.landscape.api.helpers import (
    with_computer_selection,
    with_computer,
    paginated,
    ApiResponse,
)
from canonical.landscape.api.principal import Principal
from canonical.landscape.api.schema import Array, Mapping

from canonical.landscape.model.main.computer import Computer
from canonical.landscape.model.snap.management import SnapManagement
from canonical.landscape.model.selection.selection import ComputerSelection
from canonical.landscape.environment import get_raw_config

DEFAULT_SNAP_STORE_API_URL = "https://api.snapcraft.io/v2"

CHANGE_SNAPS_SCHEMA = Schema(
    name="ChangeSnaps",
    parameters=[
        Unicode("action", doc="The action to perform for the provided snaps."),
        Date("deliver_after", optional=True),
        Integer("deliver_after_window", optional=True, default=0),
        Array("computer_ids", Integer(), doc="The numerical IDs of the computers."),
        Array(
            "snaps", Mapping(key=Unicode(), value=Union(options=[Bool(), Unicode()]))
        ),
    ],
)


@with_computer_selection(CHANGE_SNAPS_SCHEMA, "EditComputerPackages")
async def snaps_action_handler(
    request: Request,
    principal: Principal,
    selection: ComputerSelection,
    validated_data: dict,
):
    if not validated_data["computer_ids"]:
        return {
            "error": "NoComputersSelected",
            "message": "No computer IDs were provided",
        }, 400

    management = SnapManagement(selection)

    if management.has_unsupported_computers:
        unsupported_computers = ", ".join(
            str(c.id) for c in management.unsupported_computers
        )

        return {
            "error": "UnsupportedComputers",
            "message": f"Computers {unsupported_computers} do not support snap "
            "management.",
        }, 400

    action = validated_data["action"]
    changes = {action: [dict(snap) for snap in validated_data["snaps"]]}

    activity_parent = management.change_snaps(
        principal.person,
        changes=changes,
        deliver_after=validated_data.get("deliver_after"),
        deliver_after_window=validated_data["deliver_after_window"],
    )

    if not activity_parent.children_count:
        Store.of(activity_parent).remove(activity_parent)
        return {
            "error": "NoApplicableSnapActions",
            "message": f"Action '{action}' was not applicable to any provided snaps",
        }, 400

    return activity_parent.get_state(), 200


LIST_INSTALLED_SNAPS_SCHEMA = Schema(
    name="ListInstalledSnaps", doc="List installed snaps on a single computer"
)


@paginated
@with_computer(LIST_INSTALLED_SNAPS_SCHEMA, "ViewComputerPackages")
async def installed_snaps_list_handler(
    request: Request,
    principal: Principal,
    computer: Computer,
    selection: ComputerSelection,
    validated_data: dict,
    url_args: dict,
    search_args: Sequence[str] = (),
    offset=None,
    limit=None,
) -> ApiResponse:

    search = search_args[0] if search_args else None

    management = SnapManagement(selection)
    result = management.search_installed_snaps_by_name(name_fragment=search)

    count = result.count()
    result.config(limit=limit, offset=offset)

    return {
        "count": count,
        "results": [snap.serialize() for snap in result],
    }, 200


FIND_AVAILABLE_SNAPS_SCHEMA = Schema(
    name="FindAvailableSnaps",
    parameters=[Unicode("name_startswith")],
    doc="Find snaps from the public snap store",
)


@with_computer(FIND_AVAILABLE_SNAPS_SCHEMA, "ViewComputerPackages")
async def available_snaps_get_handler(
    request: Request,
    principal: Principal,
    computer: Computer,
    computer_selection: ComputerSelection,
    validated_data: dict,
    url_args: dict,
) -> ApiResponse:
    name_startswith = validated_data["name_startswith"]
    url = _get_snap_store_api_endpoint("snaps/find")
    params = {"name-startswith": name_startswith, "fields": "publisher"}
    return _get_request_snap_store(url, params)


GET_SNAP_INFO_SCHEMA = Schema(
    name="GetSnapInfo",
    doc="Get information from the public store about a particular snap",
)


@with_computer(GET_SNAP_INFO_SCHEMA, "ViewComputerPackages")
async def snap_info_get_handler(
    request: Request,
    principal: Principal,
    computer: Computer,
    computer_selection: ComputerSelection,
    validated_data: dict,
    url_args: dict,
) -> ApiResponse:
    name = url_args["snap_name"]
    url = _get_snap_store_api_endpoint(f"snaps/info/{name}")
    params = {"fields": "confinement,version,revision,publisher"}
    return _get_request_snap_store(url, params=params)


def _get_request_snap_store(url: str, params: dict = None) -> dict:
    """Make GET request to the snap store at the given url with query params"""
    headers = {"Snap-Device-Series": "16"}  # magic number requirement from Snap Store
    try:
        response = _get_response(url, params=params, headers=headers)
        if not response.ok:
            response.raise_for_status()
    except HTTPError as e:
        return {"error": e.__class__.__name__, "message": str(e)}, 400
    except ConnectionError as e:
        return {
            "error": "SnapStoreConnectionError",
            "message": f"Failed to connect to {url}.",
        }, 504
    else:
        return response.json(), 200


def _get_response(url: str, params: dict = None, headers: dict = None) -> Response:
    """Convenient for unit testing to not have to mock requests.request"""
    return get(url=url, params=params, headers=headers)


def _get_snap_store_api_endpoint(url: str) -> str:
    """Retrieve snap store endpoint from config and combine with endpoint fragment"""
    config = get_raw_config()
    snap_store_api_base = config["api"].get(
        "snap-store-api-url", DEFAULT_SNAP_STORE_API_URL
    )
    return os.path.join(snap_store_api_base, url)
