# Copyright 2005-2023 Canonical Limited. All rights reserved.
"""
User
======

The methods available here give the ability to create, read, update, and delete
Linux users on Landscape computers.
"""

from twisted.web.server import Request

from canonical.txapi.server.call import Call
from canonical.txapi.server.schema import Bool, Integer, List, Schema, Unicode
from canonical.txapi.server.method import method

from canonical.landscape.api.error import UnknownComputerError, UnknownUserError
from canonical.landscape.api.helpers import (
    paginated,
    with_computer_selection,
    with_schema,
)
from canonical.landscape.api.jwt_auth import login_required
from canonical.landscape.api.methods import AccountMethod
from canonical.landscape.api.principal import Principal
from canonical.landscape.api.routes import AccountRoute, SelectionRoute
from canonical.landscape.api.schema import Array
from canonical.landscape.api.types import ApiResponse
from canonical.landscape.model.computer.management import ComputerManagement
from canonical.landscape.model.main.account import Account, get_account_by_id
from canonical.landscape.model.main.computer import Computer
from canonical.landscape.model.main.person import get_person_by_id
from canonical.landscape.model.selection.selection import ComputerSelection
from canonical.landscape.model.users.users import UserSelection, UserManagement


def get_user_selection(
    computer_selection: ComputerSelection,
    person_id: int,
    usernames: list,
):
    person = get_person_by_id(person_id)
    management = ComputerManagement(computer_selection, person)

    users = []

    for computer in computer_selection.get_selected():
        for username in usernames:
            user = computer.get_user_by_username(username)

            if user is None:
                raise UnknownUserError(
                    f"Computer with id {computer.id} does not have user with "
                    f"username {username}"
                )

            users.append(computer.get_user_by_username(username))

    selection = UserSelection(management, users)

    return selection


def get_users_helper(computer, limit, offset):
    users = computer.get_users()
    users.config(limit=limit, offset=offset)
    return [user.get_state() for user in users]


@method
class CreateUser(AccountMethod):
    requires = "RequestCreateUser"
    route = SelectionRoute(permission=requires)
    schema = Schema(
        name="CreateUser",
        parameters=[
            List("computer_ids", Integer(), doc="The numerical IDs of the computers."),
            Unicode("username", doc="The username of the new user."),
            Unicode("name", doc="The title name of the new user."),
            Unicode("password", doc="The password of the new user."),
            Bool(
                "require_password_reset",
                optional=True,
                default=False,
                doc="Requires the user to reset their password on first login",
            ),
            Unicode(
                "primary_groupname",
                optional=True,
                doc="The group the new user will be assigned to.",
            ),
            Unicode(
                "location",
                optional=True,
                doc="The location of the new user.",
            ),
            Unicode(
                "home_phone",
                optional=True,
                doc="The home phone number of the new user.",
            ),
            Unicode(
                "work_phone",
                optional=True,
                doc="The work phone number of the new user.",
            ),
        ],
        errors=[],
        doc="Create an activity to create a Linux user on the specified computers.",
    )

    def query(self, context: ComputerSelection, call: Call) -> dict:
        person = get_person_by_id(call.principal.id)
        management = UserManagement(context, person)
        activity_group = management.create_user(
            call.args.username,
            call.args.name,
            call.args.password,
            call.args.require_password_reset,
            call.args.primary_groupname,
            call.args.location,
            call.args.home_phone,
            call.args.work_phone,
        )

        return activity_group.get_state()


@method
class EditUser(AccountMethod):
    requires = "EditUser"
    route = SelectionRoute(permission=requires)
    schema = Schema(
        name="EditUser",
        parameters=[
            List("computer_ids", Integer(), doc="The numerical IDs of the computers."),
            Unicode("username", doc="The username of an existing user."),
            Unicode(
                "name",
                optional=True,
                doc="The new title name of the existing user.",
            ),
            Unicode(
                "password",
                optional=True,
                doc="The new password for the existing user.",
            ),
            Unicode(
                "primary_groupname",
                optional=True,
                doc="The new group the existing user will be assigned to.",
            ),
            Unicode(
                "location",
                optional=True,
                doc="The new location of the existing user.",
            ),
            Unicode(
                "home_phone",
                optional=True,
                doc="The new home phone number of the existing user.",
            ),
            Unicode(
                "work_phone",
                optional=True,
                doc="The new work phone number of the existing user.",
            ),
        ],
        errors=[UnknownUserError],
        doc="Create an activity to edit information of a Linux user on the specified computers.",
    )

    def query(self, context: ComputerSelection, call: Call) -> dict:
        user_selection = get_user_selection(
            context,
            call.principal.id,
            [call.args.username],
        )
        activity_group = user_selection.edit_users(
            call.args.primary_groupname,
            call.args.name,
            call.args.password,
            call.args.location,
            call.args.work_phone,
            call.args.home_phone,
        )

        return activity_group.get_state()


@method
class GetUsers(AccountMethod):
    requires = "ViewUser"
    route = AccountRoute()
    schema = Schema(
        name="GetUsers",
        parameters=[
            Integer("computer_id", doc="The numerical ID of the computer."),
        ],
        errors=[UnknownComputerError],
        doc="Get Linux user information from the specified computer.",
    )

    def query(self, context: Account, call: Call) -> list:
        computer = context.get_computer(call.args.computer_id)
        if computer is None:
            raise UnknownComputerError(f"Unknown computer '{call.args.computer_id}'")

        return get_users_helper(computer, limit=None, offset=None)


CREATE_USER_POST_HANDLER_SCHEMA = CreateUser.schema.extend()
CREATE_USER_POST_HANDLER_SCHEMA.replace_parameter(
    "computer_ids",
    Array("computer_ids", Integer(), doc="The numerical IDs of the computers."),
)


@with_computer_selection(CREATE_USER_POST_HANDLER_SCHEMA, "RequestCreateUser")
async def create_user_post_handler(
    request: Request,
    principal: Principal,
    selection: ComputerSelection,
    validated_data: dict,
) -> ApiResponse:
    person = get_person_by_id(principal.id)
    management = UserManagement(selection, person)
    activity_group = management.create_user(
        validated_data.get("username"),
        validated_data.get("name"),
        validated_data.get("password"),
        validated_data.get("require_password_reset", False),
        validated_data.get("primary_groupname"),
        validated_data.get("location"),
        validated_data.get("home_phone"),
        validated_data.get("work_phone"),
    )

    return activity_group.get_state(), 200


EDIT_USER_HANDLER_SCHEMA = EditUser.schema.extend()
EDIT_USER_HANDLER_SCHEMA.replace_parameter(
    "computer_ids",
    Array("computer_ids", Integer(), doc="The numerical IDs of the computers."),
)


@with_computer_selection(EDIT_USER_HANDLER_SCHEMA, "EditUser")
async def edit_user_handler(
    request: Request,
    principal: Principal,
    selection: ComputerSelection,
    validated_data: dict,
) -> ApiResponse:
    try:
        user_selection = get_user_selection(
            selection,
            principal.id,
            [validated_data.get("username")],
        )
    except UnknownUserError as e:
        return {"error": "Unprocessable Content", "message": str(e)}, 422

    activity_group = user_selection.edit_users(
        validated_data.get("primary_groupname"),
        validated_data.get("name"),
        validated_data.get("password"),
        validated_data.get("location"),
        validated_data.get("work_phone"),
        validated_data.get("home_phone"),
    )

    return activity_group.get_state(), 200


@login_required
@with_schema(GetUsers.schema)
@paginated
async def users_get_handler(
    request: Request,
    principal: Principal,
    validated_data: dict,
    limit=None,
    offset=None,
):
    account = get_account_by_id(principal.account_id)
    computer_id = validated_data["computer_id"]
    computer = account.get_computer(computer_id)
    if computer is None:
        return {
            "error": "UnknownComputerError",
            "message": f"Unknown computer '{computer_id}'",
        }, 400

    users = get_users_helper(computer, limit, offset)

    result = computer.get_users()
    count = result.count()

    return {
        "count": count,
        "results": users,
    }, 200


USERS_SCHEMA = Schema(
    name="CommandUsers",
    parameters=[
        Array("computer_ids", Integer(), doc="The numerical IDs of the computers."),
        Array("usernames", Unicode(), doc="The usernames of the users to apply."),
    ],
    errors=[UnknownUserError],
    doc="Create an activity to apply an operation to users on the specified computers.",
)


@with_computer_selection(USERS_SCHEMA, "EditUser")
async def lock_user_post_handler(
    request: Request,
    principal: Principal,
    selection: ComputerSelection,
    validated_data: dict,
) -> ApiResponse:
    try:
        user_selection = get_user_selection(
            selection,
            principal.id,
            validated_data.get("usernames"),
        )
    except UnknownUserError as e:
        return {"error": "Unprocessable Content", "message": str(e)}, 422

    activity_group = user_selection.lock_users()

    return activity_group.get_state(), 200


REMOVE_USERS_SCHEMA = Schema(
    name="RemoveUser",
    parameters=[
        Array("computer_ids", Integer(), doc="The numerical IDs of the computers."),
        Array("usernames", Unicode(), doc="The usernames of the users to delete."),
        Bool(
            "delete_home",
            optional=True,
            default=False,
            doc="The user's home directory will also be deleted.",
        ),
    ],
    errors=[UnknownUserError],
    doc="Create an activity to remove a Linux user from the specified computers.",
)


@with_computer_selection(REMOVE_USERS_SCHEMA, "RemoveUser")
async def remove_user_handler(
    request: Request,
    principal: Principal,
    selection: ComputerSelection,
    validated_data: dict,
) -> ApiResponse:
    try:
        user_selection = get_user_selection(
            selection,
            principal.id,
            validated_data.get("usernames"),
        )
    except UnknownUserError as e:
        return {"error": "Unprocessable Content", "message": str(e)}, 422

    activity_group = user_selection.delete_users(validated_data.get("delete_home"))

    return activity_group.get_state(), 200


@with_computer_selection(USERS_SCHEMA, "EditUser")
async def unlock_user_post_handler(
    request: Request,
    principal: Principal,
    selection: ComputerSelection,
    validated_data: dict,
) -> ApiResponse:
    try:
        user_selection = get_user_selection(
            selection,
            principal.id,
            validated_data.get("usernames"),
        )
    except UnknownUserError as e:
        return {"error": "Unprocessable Content", "message": str(e)}, 422

    activity_group = user_selection.unlock_users()

    return activity_group.get_state(), 200
