# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from twisted.internet.defer import inlineCallbacks, returnValue

from storm.twisted.transact import transact

from canonical.lib.splice import splice

from canonical.amqp.job.runner import JobRunner, job

from canonical.landscape.model.computer.twin import ComputerTwin
from canonical.landscape.model.main.account import get_account_by_id
from canonical.landscape.model.main.person import get_person_by_id
from canonical.landscape.model.main.computer import (
    cleanup_after_removed_computers,
    get_child_computer_ids,
)
from canonical.landscape.model.selection.selection import ComputerSelection
from canonical.landscape.model.computer.management import ComputerManagement
from canonical.landscape.model.resource.store import get_resource_stores


@job
class RemoveComputersJobRunner(JobRunner):
    """Remove the selected computers from the account in batches."""

    @inlineCallbacks
    def run(self, account_id, computer_ids, person_id=None, batch_size=10):
        for ids in splice(computer_ids, batch_size):
            yield self._remove_computers(account_id, person_id, ids)
        yield self._process_computers_removal(account_id, computer_ids)
        returnValue((True, None))

    @transact
    def _remove_computers(self, account_id, person_id, computer_ids):
        """Remove computers from main and account databases."""
        account = get_account_by_id(account_id)
        person = get_person_by_id(person_id) if person_id is not None else None
        selection = ComputerSelection(account)
        selection.select_computers(computer_ids)
        management = ComputerManagement(selection, person)
        management.remove_computers()

    @transact
    def _process_computers_removal(self, account_id, computer_ids):
        """
        After computers have been removed from the main and account store,
        check account rules and create a job to remove computers from the
        resource database.
        """
        account = get_account_by_id(account_id)
        cleanup_after_removed_computers(account, computer_ids)


@job
class RemoveComputerResourcesJobRunner(JobRunner):
    """Remove the select computer-related resources from resource databases."""

    @inlineCallbacks
    def run(self, computer_ids, batch_size=10):
        """Splice computer_ids and remove them in batches."""
        for ids in splice(list(computer_ids), batch_size):
            yield self._remove_computer_resources(ids)
        returnValue((True, None))

    @transact
    def _remove_computer_resources(self, computer_ids):
        """Remove computer resource data from all resource databases."""
        for _, store in get_resource_stores():
            resource_computer_twins = store.find(
                ComputerTwin, ComputerTwin.id.is_in(computer_ids)
            )
            resource_computer_twins.remove()
