# Copyright 2005-2022 Canonical Limited. All rights reserved.

from dataclasses import dataclass

from storm.locals import Int, Unicode, DateTime, Float, AutoReload
from zope.interface import implementer

from canonical.lib.timestamp import format_datetime
from canonical.security.protect import requires, requires_advisor, PRIVATE
from canonical.storm.variables import BytesStr

from canonical.landscape.model.computer.helper import ComputerHelper
from canonical.landscape.model.resource.interfaces import IActiveProcess
from canonical.landscape.model.eventlog.management import EventLog


__all__ = ["SIGNAL_TERM", "SIGNAL_KILL", "ActiveProcess"]


__requires_prefix__ = "landscape."


SIGNAL_TERM = "TERM"
SIGNAL_KILL = "KILL"


@requires_advisor
@implementer(IActiveProcess)
class ActiveProcess(ComputerHelper):
    """Data for a single running process."""

    __storm_table__ = "active_process"

    requires("ViewActiveProcess", PRIVATE, ["id"])
    requires(
        "ViewActiveProcess",
        "EditActiveProcess",
        [
            "computer",
            "pid",
            "uid",
            "gid",
            "name",
            "state",
            "start_time",
            "vm_size",
            "cpu_utilisation",
        ],
    )

    id = Int(primary=True, allow_none=False, default=AutoReload)
    computer_id = Int(allow_none=False)
    pid = Int(allow_none=False)
    uid = Int(allow_none=False)
    gid = Int(allow_none=False)
    name = Unicode(allow_none=False)
    state = BytesStr(allow_none=False)
    start_time = DateTime(allow_none=False)
    vm_size = Int()
    cpu_utilisation = Float()

    def __init__(
        self, computer, pid, uid, gid, name, state, start_time, vm_size, cpu_utilisation
    ):
        self.computer = computer
        self.pid = pid
        self.uid = uid
        self.gid = gid
        self.name = name
        self.state = state
        self.start_time = start_time
        self.vm_size = vm_size
        self.cpu_utilisation = cpu_utilisation

    def get_state(self) -> dict:
        return {
            "id": self.id,
            "computer_id": self.computer_id,
            "pid": self.pid,
            "gid": self.gid,
            "name": self.name,
            "state": self.state,
            "start_time": format_datetime(self.start_time),
            "vm_size": self.vm_size,
            "cpu_utilisation": self.cpu_utilisation,
        }

    def serialize(self):
        process_state = self.get_state()
        process_state["state"] = process_state["state"].decode()  # stored as bytes
        return process_state

    @requires("ViewActiveProcess")
    def get_username(self):
        user = self.computer.get_user_by_uid(self.uid)
        if user:
            return user.username
        return None

    @requires("ViewActiveProcess")
    def get_groupname(self):
        group = self.computer.get_group_by_gid(self.gid)
        if group:
            return group.name
        return None

    @requires("ViewActiveProcess")
    def is_running(self):
        return self.state == b"R"

    @requires("ViewActiveProcess")
    def is_sleeping(self):
        return self.state == b"S" or self.state == b"D"

    @requires("ViewActiveProcess")
    def is_stopped(self):
        return self.state == b"T" or self.state == b"I"

    @requires("ViewActiveProcess")
    def is_dead(self):
        return self.state == b"X"

    @requires("ViewActiveProcess")
    def is_zombie(self):
        return self.state == b"Z"

    @requires("ManageActiveProcess")
    def request_end(self, parent):
        return self.signal("TERM", parent)

    @requires("ManageActiveProcess")
    def request_kill(self, parent):
        return self.signal("KILL", parent)

    @requires("ManageActiveProcess")
    def signal(self, signal: str, parent):
        from canonical.landscape.model.activity.activities import SignalProcessRequest

        child = parent.add(
            SignalProcessRequest,
            computer=self.computer,
            signal=signal,
            pid=self.pid,
            name=self.name,
            start_time=self.start_time,
        )
        event_log = EventLog(parent.account, parent.creator)
        event_log.create_log_entry(
            self.computer, "Activity created: " + child.info.get_summary()
        )
        return child
