# Copyright 2005-2023 Canonical Limited. All rights reserved.

import abc
import itertools

from canonical.lib.hardware import HardwareFormatter

from canonical.landscape.model.resource.hardware_info_parser.key_parser import (
    deindex_hardware,
    get_devices_info,
    get_value,
    group_by_key,
    group_by_resource_key,
    remove_keys,
    split_key,
    NOT_AVAILABLE,
)


class HardwareComponentInfoParser(abc.ABC):
    """The base class for individual harware component parsers.

    All child classes should override the appropriate `filters` and `matches`
    attributes and implement their own `parse()` method.
    """

    filters = []
    matches = []

    @abc.abstractmethod
    def parse(hardware_info: list):
        ...


class CPUInfoParser(HardwareComponentInfoParser):
    filters = []
    matches = ["cpu"]

    @classmethod
    def parse(cls, hardware_info: list) -> dict:
        """Return CPU information."""
        cpu_info = []
        cpus = group_by_resource_key(hardware_info, "cpu", ignore=["logicalcpu"])

        for cpu in cpus:
            architecture = get_value(cpu, "width")
            cache = cls.parse_cache(cpu)
            clock_speed = cls.parse_clock_speed(cpu)
            flags = cls.parse_flags(cpu)
            model = get_value(cpu, "product")
            vendor = get_value(cpu, "vendor")

            cpu_info.append(
                {
                    "architecture": architecture,
                    "cache": cache,
                    "clock_speed": clock_speed,
                    "flags": flags,
                    "model": model,
                    "vendor": vendor,
                }
            )

        if not cpu_info:
            cpu_info.append(
                {
                    "vendor": NOT_AVAILABLE,
                    "model": NOT_AVAILABLE,
                    "flags": [],
                    "architecture": NOT_AVAILABLE,
                    "clock_speed": NOT_AVAILABLE,
                    "cache": {},
                }
            )

        return cpu_info

    @staticmethod
    def parse_cache(cpu: tuple) -> dict:
        cache = {}
        cache_key = group_by_key(cpu, "cache")

        for level_key in ("slot", "description"):
            # Do the best to find cache level information using
            # some preferred keys
            cache_levels = group_by_key(cache_key, level_key)

            if cache_levels:
                cache_sizes = group_by_key(cache_key, "size")

                if cache_sizes:
                    cache_levels = [str_val for key, str_val, int_val in cache_levels]
                    cache_sizes = [str_val for key, str_val, int_val in cache_sizes]
                    cache_sizes = [
                        HardwareFormatter("cpucache", x).format() for x in cache_sizes
                    ]
                    cache = {
                        level: size for level, size in zip(cache_levels, cache_sizes)
                    }

        return cache

    @staticmethod
    def parse_clock_speed(cpu: tuple) -> str:
        clock_speed = NOT_AVAILABLE

        for key in ("capacity", "clock", "size"):
            clock_speed_key = group_by_key(cpu, key, ignore="cache")

            if clock_speed_key:
                clock_speed = clock_speed_key[0][1]
                clock_speed = HardwareFormatter("clockspeed", clock_speed).format()

                break

        return clock_speed

    @staticmethod
    def parse_flags(cpu: tuple) -> list:
        flags = []
        flag_key = group_by_key(cpu, "capabilities", ignore="cache")

        for key, str_value, int_value in flag_key:
            flag = str_value.split(" ")

            if len(flag) > 1:
                # We have flag code and description
                flags.append({"code": flag[0], "title": " ".join(flag[1:])})
            else:
                # We don't have description. Use code as title.
                flags.append({"code": flag[0], "title": flag[0]})

        return flags


class DisplayInfoParser(HardwareComponentInfoParser):
    filters = []
    matches = ["display"]

    @staticmethod
    def parse(hardware_info: list) -> dict:
        """Return display information"""
        display = group_by_key(hardware_info, "display")
        vendor = get_value(display, "vendor", start=False)
        product = get_value(display, "product", start=False)

        display_info = {"model": product, "vendor": vendor}

        return display_info


class MemoryInfoParser(HardwareComponentInfoParser):
    filters = []
    matches = ["^memory"]

    @staticmethod
    def parse(hardware_info: list) -> dict:
        """Get memory information.

        Size information can be retrieved from many different keys, depending
        on the stored information, such as:

        memory.size
        memory.bank.size
        memory:0.size
        memory:0.bank:0.size

        The algorithm first looks for a summary record and if not found it sums
        up each bank of RAM to find the total.
        """
        computer_memory = group_by_key(hardware_info, "memory", start=True)
        # We are going to aggregate over all devices, we don't care what
        # number they are and the numbers just get in the way.
        computer_memory = list(map(deindex_hardware, computer_memory))

        memory_groups = group_by_resource_key(hardware_info, "memory")

        sizes = []

        # Try to get the size attribute from "memory" entries.
        for group in memory_groups:
            size = get_value(group, "size", int_value=True, default=None)

            if size is not None:
                sizes.append(size)
            else:
                # There's no "size" attribute, use "memory.bank:X.size"
                # instead.
                bank_groups = group_by_resource_key(group, "bank")

                for group in bank_groups:
                    bank_size = get_value(group, "size", int_value=True, default=None)

                    if bank_size is not None:
                        sizes.append(bank_size)

        total = sum(sizes)

        if total > 0:
            size = HardwareFormatter("memory", total).format() + " RAM"
        else:
            size = NOT_AVAILABLE

        return {"size": size}


class MultimediaInfoParser(HardwareComponentInfoParser):
    filters = []
    matches = ["multimedia"]

    @staticmethod
    def parse(hardware_info: list) -> dict:
        """Return multimedia information."""
        multimedia = group_by_key(hardware_info, "multimedia")
        vendor = get_value(multimedia, "vendor", start=False)
        model = get_value(multimedia, "product", start=False)

        data = {"model": model, "vendor": vendor}

        return data


class NetworkInfoParser(HardwareComponentInfoParser):
    filters = []
    matches = ["^pci.*network"]

    @staticmethod
    def parse(hardware_info: list) -> list:
        """Return network information"""
        network_info = []

        for group in group_by_resource_key(hardware_info, "network"):
            vendor = get_value(group, "vendor")
            product = get_value(group, "product")
            description = get_value(group, "description")
            mac = get_value(group, "serial", start=False)
            ip = get_value(group, "configuration.ip", start=False)
            data = {
                "description": description,
                "ip": ip,
                "mac": mac,
                "product": product,
                "vendor": vendor,
            }

            network_info.append(data)

        return network_info or NOT_AVAILABLE


class PCIInfoParser(HardwareComponentInfoParser):
    filters = ["multimedia", "usb"]
    matches = ["pci"]

    @classmethod
    def parse(cls, hardware_info: list) -> list:
        """Return PCI information."""
        pci_info = []

        for group in group_by_resource_key(hardware_info, "pci.pci"):
            device_keys = cls.get_device_keys(group)
            attached_devices = cls.get_attached_devices(group, device_keys)

            bus_info = {
                "attached_devices": attached_devices,
                "description": get_value(group, "description"),
                "model": get_value(group, "product"),
                "vendor": get_value(group, "vendor"),
            }

            pci_info.append(bus_info)

        return pci_info

    @staticmethod
    def get_attached_devices(pci_group: list, device_keys: set) -> list:
        attached_devices = []

        for device_key in device_keys:
            [device_group] = group_by_resource_key(pci_group, device_key, start=True)
            attached_devices.append(
                {
                    "vendor": get_value(device_group, "vendor"),
                    "model": get_value(device_group, "product"),
                    "description": get_value(device_group, "description"),
                }
            )

        return attached_devices

    @staticmethod
    def get_device_keys(pci_group: list) -> set:
        device_keys = set()

        for info in pci_group:
            key = info[0]
            split = key.split(".", 1)

            if len(split) == 2:
                device_key = split[0]

                # Ingnore bus information sections
                if device_key not in ("resources", "configuration"):
                    device_keys.add(device_key)

        return device_keys


class SCSIInfoParser(HardwareComponentInfoParser):
    filters = []
    matches = ["scsi"]

    @staticmethod
    def parse(hardware_info: list) -> list:
        """Return SCSI disk info."""
        return get_devices_info(hardware_info, "scsi.disk")


class StorageInfoParser(HardwareComponentInfoParser):
    filters = []
    matches = ["disk"]

    @staticmethod
    def parse(hardware_info: list) -> list:
        """Return storage information"""
        storage_info = []

        disk_info = group_by_resource_key(hardware_info, "disk")
        scsi_info = group_by_resource_key(hardware_info, r"scsi(:\d+)?.virtio\d*")
        nvme_info = group_by_resource_key(hardware_info, r"nvme\d*")

        for group in itertools.chain(disk_info, scsi_info, nvme_info):
            vendor = get_value(group, "vendor")
            product = get_value(group, "product")
            description = get_value(group, "description")
            size = get_value(group, "size", default=None) or get_value(
                group, "namespace.size"
            )
            size = HardwareFormatter("disk", size).format()

            volume_group = group_by_resource_key(group, "volume")
            partitions = []

            for volume in volume_group:
                partition_description = get_value(volume, "description")
                partition_size = get_value(volume, "size")
                partition_size = HardwareFormatter("disk", partition_size).format()
                partition_filesystem = get_value(volume, "configuration.filesystem")
                data = {
                    "description": partition_description,
                    "filesystem": partition_filesystem,
                    "size": partition_size,
                }

                partitions.append(data)

            data = {
                "description": description,
                "partitions": partitions,
                "product": product,
                "size": size,
                "vendor": vendor,
            }

            storage_info.append(data)

        return storage_info or NOT_AVAILABLE


class SystemInfoParser(HardwareComponentInfoParser):
    filters = []
    matches = ["core.", "^configuration"]

    @staticmethod
    def parse(hardware_info: list) -> dict:
        """Return system information."""
        vendor = get_value(hardware_info, "core.vendor")
        serial = get_value(hardware_info, "serial")
        model = get_value(hardware_info, "product")
        chassis = get_value(hardware_info, "configuration.chassis")

        # BIOS info is pulled from firmware
        bios_info = group_by_key(hardware_info, "firmware", start=True)
        bios_vendor = get_value(bios_info, "firmware.vendor")
        bios_version = get_value(bios_info, "firmware.version")
        bios_date = get_value(bios_info, "firmware.date")

        bios_capabilities = []
        capabilities = group_by_key(bios_info, "capabilities")

        for key, str_value, int_value in capabilities:
            flag = str_value.split(" ")

            if len(flag) > 1:
                # We have flag code and description
                bios_capabilities.append({"code": flag[0], "title": " ".join(flag[1:])})
            else:
                # We don't have description. Use code as title.
                bios_capabilities.append({"code": flag[0], "title": flag[0]})

        system_info = {
            "bios_capabilities": bios_capabilities,
            "bios_date": bios_date,
            "bios_vendor": bios_vendor,
            "bios_version": bios_version,
            "chassis": chassis,
            "model": model,
            "serial": serial,
            "vendor": vendor,
        }

        return system_info


class USBInfoParser(HardwareComponentInfoParser):
    filters = []
    matches = ["usb"]

    @staticmethod
    def parse(hardware_info: list) -> list:
        """Return USB info."""
        return get_devices_info(hardware_info, "pci.usb")
