# Copyright 2005-2023 Canonical Limited. All rights reserved.

import logging
import re
import sys

from canonical.landscape.model.resource.hardware_info_parser.hardware_component_info_parser import (
    CPUInfoParser,
    DisplayInfoParser,
    MemoryInfoParser,
    MultimediaInfoParser,
    NetworkInfoParser,
    PCIInfoParser,
    SCSIInfoParser,
    StorageInfoParser,
    SystemInfoParser,
    USBInfoParser,
)
from canonical.landscape.model.resource.hardware_info_parser.key_parser import (
    NOT_AVAILABLE,
)


class HardwareInfoParser:
    _hardware_classes = {
        "cpu": CPUInfoParser,
        "display": DisplayInfoParser,
        "memory": MemoryInfoParser,
        "multimedia": MultimediaInfoParser,
        "network": NetworkInfoParser,
        "pci": PCIInfoParser,
        "scsi": SCSIInfoParser,
        "storage": StorageInfoParser,
        "system": SystemInfoParser,
        "usb": USBInfoParser,
    }

    def __init__(self, computer_id, hardware_info):
        self.computer_id = computer_id
        self.hardware = self._get_hardware(hardware_info)
        self._validate_hardware(hardware_info)

    def _get_hardware(self, hardware_info):
        """
        Collect and return all parsed hardware information. If colletion of
        a class of hardware fails, log the error and move on and fill in the
        other class data.
        """
        parsed_hardware = {}

        for hardware_class, parser in self._hardware_classes.items():
            try:
                parsed_hardware[hardware_class] = parser.parse(hardware_info)
            except Exception as error:
                # Catch any exceptions which could range from invalid index
                # errors to divsion by zero
                logging.error(
                    f"Hardware Data: Failure parsing {hardware_class} data on "
                    f"computer {self.computer_id}: {str(error)} "
                    f"(line:{sys.exc_info()[-1].tb_next.tb_lineno})"
                )

                parsed_hardware[hardware_class] = []

        return parsed_hardware

    def _validate_hardware(self, hardware_info):
        """
        Quick health check to ensure we have extracted some valid hardware
        data appropriately. If we have significant gaps in data, we will
        log some warnings in landscape logs.
        """
        if len(hardware_info) == 0:
            return

        # Collect counts of potential key matches from lshw.
        # We have a problem if we have a count of potential key matches but
        # no hardware data found for that class.
        counts = dict.fromkeys(self._hardware_classes, 0)

        all_hardware = [key for key, str_val, int_val in hardware_info]

        # Count up potential hardware class matches from all_hardware
        for hardware_name in all_hardware:
            for hardware_class, parser in self._hardware_classes.items():
                # Check match patterns for a potential match for each hardware
                # class. Tally each match and move to the next hardware_name.
                if any(re.search(regex, hardware_name) for regex in parser.filters):
                    continue

                if any(re.search(regex, hardware_name) for regex in parser.matches):
                    counts[hardware_class] += 1
                    break

        warn_hardware_class_data = []
        for hardware_class, devices in list(self.hardware.items()):
            valid_data = False
            if isinstance(devices, dict):
                for attribute, value in list(devices.items()):
                    if value != NOT_AVAILABLE:
                        valid_data = True
            elif isinstance(devices, list):
                if len(devices) > 0:
                    valid_data = True
            else:
                if devices != NOT_AVAILABLE:
                    valid_data = True

            # Only append warning if we don't see any valid_data, but
            # we have potential matches in hardware_class_data
            if not valid_data:
                if counts[hardware_class] > 0:
                    warn_hardware_class_data.append(hardware_class)

        if len(warn_hardware_class_data) > 0:
            logging.warn(
                "Hardware Data: Could not parse %s data for computer %s. "
                "There are %s attributes seen from lshw which could be "
                "related to this data."
                % (
                    ", ".join(warn_hardware_class_data),
                    self.computer_id,
                    ", ".join(
                        "%d %s" % (counts[key], key) for key in warn_hardware_class_data
                    ),
                )
            )
