import json
import logging
import os
import sys
from typing import Tuple

import requests
from twisted.internet import defer, reactor, task
from twisted.web import server, resource

from canonical.landscape.api.error import (
    SecretsAuthenticationError,
    SecretNotFoundError,
    SecretsSealedError,
)
from canonical.landscape.api.secrets_client import HTTPVaultClient
from canonical.landscape.api.types import ApiResponse
from canonical.landscape.application import setup_logging
from canonical.landscape.model.main.account import DEFAULT_STANDALONE_ACCOUNT_NAME
from canonical.landscape.setup import load_config


REFRESH_INTERVAL_SECONDS = 60 * 60
SERVICE_NAME = "landscape-secrets-service"
TTL = "4h"


def create_policy(client: HTTPVaultClient) -> None:
    """Creates a policy to allow the token to only access the store for Landscape"""
    try:
        policy = client.get_policy(SERVICE_NAME)
    except SecretNotFoundError:
        logging.info("Creating policy..")
        policy = """
        path "{}/*" {{
        capabilities = ["create", "read", "update", "delete", "list"]
        }}
        path "sys/mounts/{}/*" {{
        capabilities = ["read", "list"]
        }}
        """.format(
            DEFAULT_STANDALONE_ACCOUNT_NAME, DEFAULT_STANDALONE_ACCOUNT_NAME
        )
        client.create_policy(SERVICE_NAME, policy)


def get_token(client: HTTPVaultClient) -> str:
    """Gets a new vault token with the specified policy and TTL"""
    policy_name = SERVICE_NAME
    return client.create_orphan_token(TTL, policy_name)


def handle_error(twisted_error_obj):
    """Log the error and stop the service"""
    logging.error(twisted_error_obj.getBriefTraceback())
    os._exit(1)  # We need to kill all twisted processes on error


class Secret(resource.Resource):
    isLeaf = True

    def __init__(self, secrets_url: str):
        super().__init__()

        self.secrets_url = secrets_url
        self.token = None

    def render_GET(self, request: server.Request):
        return json.dumps({"token": self.token}).encode()

    def render_POST(self, request):
        status, body = self.handle_POST(request)

        request.setHeader("Content-Type", "application/json")
        request.setResponseCode(status)
        request.write(json.dumps(body).encode())
        request.finish()

        return server.NOT_DONE_YET

    def handle_POST(self, request: server.Request) -> ApiResponse:
        if self.token:
            return 400, {
                "error": "SecretsAuthenticationError",
                "message": "Token has already been set. To use a different token, please restart the service.",
            }

        try:
            content = request.content.read().decode()
        except UnicodeDecodeError:
            return 400, {
                "error": "UnicodeDecodeError",
                "message": "Request body is not valid UTF-8",
            }

        data = {}

        if content:
            try:
                data = json.loads(content)
            except json.decoder.JSONDecodeError:
                return 400, {
                    "error": "JSONDecodeError",
                    "message": "Request body is not valid JSON",
                }

        token = data.get("token")

        if not token:
            return 400, {
                "error": "MissingParameterError",
                "message": "Missing field 'token' in request body",
            }

        try:
            client = HTTPVaultClient(
                DEFAULT_STANDALONE_ACCOUNT_NAME,
                token=token,
                url=self.secrets_url,
            )
        except requests.exceptions.ConnectionError:
            return 503, {
                "error": "ConnectionError",
                "message": "Please check that Vault is running and the address is correct!",
            }
        except SecretsAuthenticationError as error:
            return 503, {
                "error": "SecretsAuthenticationError",
                "message": "Please check that the access token is correct!",
                "mes": str(error),
            }
        except SecretsSealedError:
            return 503, {
                "error": "SecretsSealedError",
                "message": "Vault is sealed! Please unseal!",
            }

        create_policy(client)
        self.token = get_token(client)
        self.create_task(client)

        return 200, {"message": "Token set"}

    def create_task(
        self, client: HTTPVaultClient
    ) -> Tuple[defer.Deferred, task.LoopingCall]:
        scheduler = task.LoopingCall(self.refresh_token, client)
        deferred = scheduler.start(REFRESH_INTERVAL_SECONDS)
        deferred.addErrback(handle_error)

        return deferred, scheduler

    def refresh_token(self, client: HTTPVaultClient) -> None:
        """Gets a new token and expires the previous one"""
        logging.info("Refreshing token..")
        prev = self.token
        self.token = get_token(client)
        client.revoke_token(prev)


def load_secrets_config() -> Tuple[str, int]:
    config = load_config("secrets")
    secrets_url = config.get("secrets", "secrets-url")
    service_url = config.get("secrets", "secrets-service-url")

    if service_url:
        try:
            service_port = int(service_url.split(":")[-1])
        except ValueError:
            logging.error("Port must be specified in service url Landscape config!")
            sys.exit(1)
    else:
        logging.error("Service url must be specified in the Landscape config!")
        sys.exit(1)

    return secrets_url, service_port


def run() -> None:
    setup_logging("landscape-secrets-service", always_log_to_stdout=True)
    logger = logging.getLogger(SERVICE_NAME)

    logging.info("Starting service..")
    secrets_url, service_port = load_secrets_config()

    logging.info(f"Using port {service_port}")
    site = server.Site(Secret(secrets_url))

    # Only allow access from localhost
    reactor.listenTCP(service_port, site, interface="127.0.0.1")

    reactor.run()
