# Copyright 2005-2024 Canonical Limited.  All rights reserved.

"""
Package Profiles
================

A package profile is a set of package constraints that can be applied to
managed computers.

The methods available are related to package profile management. The user can
search, create, copy, read, update and delete profiles associated with the
account.
"""

from copy import deepcopy
from dataclasses import asdict, dataclass

from typing import Dict, List as ListType, Mapping, Sequence, Union

from twisted.web.server import Request

from zope.security.interfaces import Unauthorized

from canonical.txapi.server.method import method
from canonical.txapi.server.schema import Bool, Schema, Unicode, Integer, List

from canonical.landscape.api.error import (
    UnknownComputerError,
    InvalidParameterCombinationError,
    MissingParameterError,
)
from canonical.landscape.api.jwt_auth import login_required
from canonical.landscape.api.helpers import (
    EMPTY_BODY,
    add_search_query,
    paginated,
    with_computer,
    with_management_class,
    with_schema,
)
from canonical.landscape.api.methods import AccountMethod, TagAssociationMethod
from canonical.landscape.api.principal import Principal
from canonical.landscape.api.routes import ManagementRoute, SubjectRoute
from canonical.landscape.api.schema import Array, UnicodeTitle, Mapping as ApiMapping
from canonical.landscape.api.types import ApiResponse

from canonical.landscape.model.computer.collection import ComputerCollection
from canonical.landscape.model.main.account import DEFAULT_ACCESS_CONTEXT_NAME
from canonical.landscape.model.main.computer import Computer
from canonical.landscape.model.main.errors import UnknownAccessContextError
from canonical.landscape.model.main.person import get_person_by_id
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.tag import InvalidTagError, get_tag_names, get_tags
from canonical.landscape.model.package.package import (
    InvalidPackageConstraintError,
    InvalidPackageConstraintTypeError,
    build_relation_type,
    build_relation_value,
    parse_constraint,
)
from canonical.landscape.model.package.profile import (
    PackageProfileConstraint,
    PackageProfileError,
    PackageProfilesManagement,
    UnknownPackageProfileNameError,
    DuplicatePackageProfileError,
    NoFoundPackagesError,
    InvalidPackageProfileMaterialError,
    EmptyPackageProfileError,
    InvalidPackageProfileNameError,
    create_profile_from_computer_selection,
    parse_package_profile_material,
)
from canonical.landscape.model.selection.selection import ComputerSelection


__all__ = [
    "GetPackageProfiles",
    "CopyPackageProfile",
    "CreatePackageProfile",
    "RemovePackageProfile",
    "AssociatePackageProfile",
    "DisassociatePackageProfile",
    "EditPackageProfile",
    "create_package_profile_constraints_handler",
    "create_package_profile_handler",
    "delete_package_profile_constraints_handler",
    "get_computer_package_profile_handler",
    "list_package_profile_constraints_handler",
    "list_package_profiles_handler",
    "parse_package_profile_material_handler",
    "update_package_profile_handler",
    "update_package_profile_constraint_handler",
]


@method
class GetPackageProfiles(AccountMethod):
    """
    Get the details of all Package Profiles defined in the account.

    - **names**: Specific profile names to limit results. (optional)

    For example, the following command limits the results to return the single
    package profile named "mysqlprofile" in this account::

        ?action=GetPackageProfiles&names.1=mysqlprofile

    The method returns a JSON representation of the package profiles. It
    includes a list of computers which are constrained by this profile as well
    as a list of constrained computers which are non-compliant::

        [{"id": 2,
          "creation_time": "2012-03-06T18:52:09Z",
          "computers": {"constrained": [4, 13, 79, 945],
                        "non-compliant": [13]},
          "modification_time": "2012-03-06T18:52:09Z",
          "description": "Package profile to lock down mysql to >= 5.1.0",
          "version": "2",
          "tags": ["server", "desktop"],
          "constraints": [("depends", "mysql >= 5.1.0"),
                          ("conflicts", "python-mysqldb < 1.2.2")],
          "name": "mysqlprofile",
          "title": "MySQL Profile"}]
    """

    requires = "ViewAccount"
    schema = Schema(
        name="GetPackageProfiles",
        parameters=[
            List(
                "names",
                Unicode(),
                optional=True,
                doc="A list of package profile names to limit the result.",
            )
        ],
        errors=[],
        doc="""Get the details of all Package Profiles defined in the account.
            """,
    )

    route = ManagementRoute(factory=PackageProfilesManagement)

    def query(self, context, call):
        """
        Return a list of C{PackageProfile}s
        """
        result = context.get_package_profiles(names=call.args.names or None)
        return [profile.get_state() for profile in result]


@method
class CopyPackageProfile(AccountMethod):
    """
    Copy an existing package profile to a package profile with a new name.

    - **name**: A name of the existing package profile to copy.
    - **destination_name**: The profile name of the copied package profile.
      (optional)
    - **title**: The title of the copied package profile.  By default the
      title of the source profile is used. (optional)
    - **description**: The descrption of the copied package profile.  By
      default the descrption of the source profile is used. (optional)
    - **access_group**: An optional name of the access group to copy the
      profile to. Defaults to the origin's access group.

    For example, the following command copies the package profile
    'mysqlprofile' to 'copiedprofilename' within registered with the account::

        ?action=CopyPackageProfile&name=mysqlprofile
            &destination_name=copiedprofilename

    The method returns a JSON representation of the copied package profile::

        [{"id": 3,
          "creation_time": "2012-03-06T18:52:09Z",
          "computers": {"constrained": [4, 13, 79, 945],
                        "non-compliant": [13]},
          "modification_time": "2012-03-06T18:52:09Z",
          "description": "Package profile to lock down mysql to >= 5.1.0",
          "version": "2",
          "constraints": [("depends", "mysql >= 5.1.0"),
                          ("conflicts", "python-mysqldb < 1.2.2")],
          "name": "copiedprofilename"}]

    The following errors may be raised:

    - **UnknownPackageProfileName**: source package profile name does not
      exist.
    - **DuplicatePackageProfile**: destination package profile name already
      exists.
    - **UnknownAccessGroup**: The given access group is not known.
    - **Unauthorised**: The person is not authorised to copy package
      profiles in the given access group.
    """

    requires = "ViewAccount"
    schema = Schema(
        name="CopyPackageProfile",
        parameters=[
            Unicode("name", doc="A name of the existing package profile to copy."),
            Unicode(
                "destination_name",
                optional=True,
                doc="The profile name of the copied package profile.",
            ),
            Unicode(
                "title",
                optional=True,
                doc=(
                    "A title for the new profile. If not specified, the"
                    " title of the source profile is used."
                ),
            ),
            Unicode(
                "description",
                optional=True,
                doc=(
                    "A description for the new profile. If not specified,"
                    " the title of the source profile is used."
                ),
            ),
            Unicode(
                "access_group",
                optional=True,
                doc=(
                    "Name of the access group to copy the "
                    "profile to. Defaults to the origin's access group."
                ),
            ),
        ],
        errors=[
            UnknownPackageProfileNameError,
            DuplicatePackageProfileError,
            UnknownAccessContextError,
            Unauthorized,
        ],
        doc="""
    Copy an existing package profile to a package profile with a new name and
    optionally a different title and description.
    """,
    )

    route = SubjectRoute(
        factory=PackageProfilesManagement,
        getter="get_package_profile_by_name",
        identifier_arg="name",
        unknown_error_factory=UnknownPackageProfileNameError,
    )

    def query(self, context, call):
        """
        Copy a C{PackageProfile} and return the copied C{PackageProfile} state.
        """
        # We don't technically need the source profile object to copy it, since
        # the management method takes the source profile's name. We make the
        # profile the context only for improved permission checking.
        management = self.get_management(PackageProfilesManagement, call)
        copy = management.copy_package_profile(
            call.args.name,
            name=call.args.destination_name,
            title=call.args.title,
            description=call.args.description,
            access_context_name=call.args.access_group,
        )
        return copy.get_state()


@method
class CreatePackageProfile(AccountMethod):
    """
    Create a package profile.

    - **title**: The title of the package profile to create.
    - **description**: The description of the new profile.
    - **source_computer_id**: A computer ID to find a computer which will be
      used as the basis of the package profile. (optional)
    - **material**: Package data in the format of 'dpkg --get-selections' or
      CSV (as exported by Landscape). (optional)
    - **constraints**: A list of constraints to populate the package
      profile. (optional)
    - **access_group**: An optional name of the access group to create the
      profile into.

    source_computer_id and material are mutually exclusive.

    For example, the following command creates a package profile 'newprofile'
    based on the packages installed on computer 500::

        ?action=CreatePackageProfile&title=New%20profile
            &description=New%20profile
            &source_computer_id=500

    The method returns a JSON representation of the created package profile::

        [{"id": 3,
          "creation_time": "2012-03-06T18:52:09Z",
          "computers": {"constrained": [4, 13, 79, 945],
                        "non-compliant": [13]},
          "modification_time": "2012-03-06T18:52:09Z",
          "description": "New profile",
          "version": "1",
          "constraints": [("depends", "mysql >= 5.1.0"),
                          ("conflicts", "python-mysqldb < 1.2.2")],
          "name": "new-profile",
          "title": "New profile"}]

    The following errors may be raised:

    - **InvalidPackageProfileName**: The name specified is not valid.
    - **InvalidParameterCombination**: source_computer_id and material were
      both passed.
    - **UnknownComputer**: An invalid source_computer_id was passed.
    - **InvalidPackageProfileMaterial**: The material passed could not be
      parsed.
    - **NoFoundPackages**: The specified computer did not have any package data
    - **UnknownAccessGroup**: The given access group is not known.
    - **Unauthorised**: The person is not authorised to create package
      profiles in the given access group.  profile with no constraints. At
      least one valid constraint must be provided.

    """

    schema = Schema(
        name="CreatePackageProfile",
        parameters=[
            UnicodeTitle("title", doc="The title of the package profile to create."),
            Unicode("description", doc="The description of the new profile."),
            Integer(
                "source_computer_id",
                optional=True,
                doc=(
                    "A computer ID to find a computer which will be used "
                    "as the basis of the package profile."
                ),
            ),
            Unicode(
                "material",
                optional=True,
                doc=(
                    "Package data in the format of "
                    "'dpkg --get-selections' or CSV (as exported by "
                    "Landscape)."
                ),
            ),
            List(
                "constraints",
                Unicode(),
                optional=True,
                doc=(
                    "Alternative to material, constraint specifications "
                    'in the form of "depends packagename" or '
                    '"conflicts packagename < 1.0".'
                ),
            ),
            Unicode(
                "access_group",
                optional=True,
                default=DEFAULT_ACCESS_CONTEXT_NAME,
                doc=("Optional name of the access group to create the profile into"),
            ),
        ],
        errors=[
            InvalidPackageProfileNameError,
            InvalidParameterCombinationError,
            EmptyPackageProfileError,
            UnknownComputerError,
            InvalidPackageProfileMaterialError,
            NoFoundPackagesError,
            UnknownAccessContextError,
            InvalidPackageConstraintError,
            Unauthorized,
            InvalidPackageConstraintTypeError,
            DuplicatePackageProfileError,
        ],
        doc="""
    Create a package profile.

    source_computer_id and material are mutually exclusive.
    """,
    )

    requires = "ViewAccount"

    route = ManagementRoute(factory=PackageProfilesManagement)

    def query(self, context, call):
        """
        Create a package profile and return its state.
        """
        if call.args.source_computer_id:
            collection = ComputerCollection(
                get_main_store(),
                account_id=call.principal.account_id,
                person_id=call.principal.id,
                permission="landscape.ViewComputerPackages",
            )
            collection = collection.with_computer_ids([call.args.source_computer_id])
            computer = collection.find().one()
            if computer is None:
                raise UnknownComputerError(
                    "Unknown computer %s" % (call.args.source_computer_id,)
                )
        else:
            computer = None

        profile = context.create_package_profile(
            call.args.title,
            call.args.description,
            material=call.args.material,
            computer=computer,
            access_context_name=call.args.access_group,
        )

        person = get_person_by_id(call.principal.id)
        profile.modify_package_constraints(add_constraints=call.args.constraints)

        profile.on_create(person)
        return profile.get_state()


@method
class RemovePackageProfile(AccountMethod):
    """
    Remove a package profile, given its name.

    Example of a valid request that removes profile 'foobar'::

        ?action=RemovePackageProfile&name=foobar

    The following errors may be raised:

    - **UnknownPackageProfileName**: No package profile with the specified name
      exists.
    """

    requires = "EditPackageProfile"

    schema = Schema(
        name="RemovePackageProfile",
        parameters=[Unicode("name", doc="The name of the package profile to remove.")],
        errors=[UnknownPackageProfileNameError],
        doc="""Remove a package profile, given its name.""",
    )

    route = SubjectRoute(
        factory=PackageProfilesManagement,
        getter="get_package_profile_by_name",
        identifier_arg="name",
        unknown_error_factory=UnknownPackageProfileNameError,
    )

    def query(self, context, call):
        """Remove a package profile."""
        profile = context  # For readability
        # Since the delete logic is already on the management class, let's
        # reuse it.
        management = self.get_management(PackageProfilesManagement, call)
        management.remove_package_profiles([profile])
        return {}


@method
class AssociatePackageProfile(TagAssociationMethod):
    """
    Associate a package profile to computers with specific tags or to all
    computers.

    - **name**: Name of the package profile.
    - **tags.#**: List of tag names to associate to the profile.
    - **all_computers**: true if the profile should be associated to all
      computers. This parameter is *optional* and defaults to
      *false*. Individual tags associated to the profile will remain, but they
      will only be effective if the "all_computers" flag is later disabled.

    **tags.#** and **all_computers=true** are mutually exclusive.

    Example of a valid request that associates the profile 'foobar' with the
    tags 'desktop' and 'laptop'::

        ?action=AssociatePackageProfile&name=foobar
            &tags.1=desktop
            &tags.2=laptop
            &all_computers=false

    The following errors may be raised:

    - **UnknownPackageProfileName**: No package profile with the specified name
      exists.

    - **InvalidParameterCombination**: Some tags were specified and a 'true'
      value for 'all_computers' was passed.

    The state of the profile will be returned.
    """

    requires = "EditPackageProfile"

    schema = TagAssociationMethod.schema.extend(
        name="AssociatePackageProfile",
        errors=[UnknownPackageProfileNameError],
        doc="""
    Associate a package profile to computers with specific tags or to all
    computers. If a tag with a given name doesn't exist, it will be
    automatically created.

    An 'all_computers' value of 'true' and a list of tags are mutually
    exclusive. Only one or the other may be passed.""",
    )

    route = SubjectRoute(
        factory=PackageProfilesManagement,
        getter="get_package_profile_by_name",
        identifier_arg="name",
        unknown_error_factory=UnknownPackageProfileNameError,
    )


@method
class DisassociatePackageProfile(TagAssociationMethod):
    """
    Disassociate a package profile from computers with specified tags or from
    all computers.

    If **all_computers=true**, the profile will be unflagged as applying to all
    computers, but will still be enabled for computers which have specific tags
    associated with it.

    - **name**: Name of the package profile.
    - **tags.#**: List of tag names to disassociate from the profile.
    - **all_computers**: If true, the profile will only remain enabled for
      computers with tags associated to the profile. This parameter is
      *optional* and defaults to *false*.

    **tags.#** and **all_computers=true** are mutually exclusive.

    Example of a valid request::

        ?action=DisassociatePackageProfile&name=foobar
            &tags.1=desktop&tags.2=laptop

    The following errors may be raised:

    - **UnknownPackageProfileName**: No package profile with the specified name
      exists.
    - **UnknownTag**: No tag with the specified name exists.
    - **InvalidParameterCombination**: Some tags were specified and a 'true'
      value for 'all_computers' was passed.

    The state of the profile will be returned.
    """

    requires = "EditPackageProfile"

    associate = False

    schema = TagAssociationMethod.schema.extend(
        name="DisassociatePackageProfile",
        errors=[UnknownPackageProfileNameError],
        doc="""Disassociate package profile from computers with specified
            tags or from all computers.""",
    )

    route = SubjectRoute(
        factory=PackageProfilesManagement,
        getter="get_package_profile_by_name",
        identifier_arg="name",
        unknown_error_factory=UnknownPackageProfileNameError,
    )


@method
class EditPackageProfile(AccountMethod):
    """
    Add or remove constraints related to a package profile. Constraints can be
    dependencies or conflicts.

    - **name**: Name of the package profile.
    - **title**: The new title of the package profile.
    - **add_constraints.#**: Constraint specifications to add in the form of
      "depends packagename" or "conflicts packagename < 1.0".
    - **remove_constraints.#**: Constraint specifications to remove in the form
      of "depends packagename" or "conflicts packagename < 1.0".

    Example of a valid request::

        ?action=EditPackageProfile&name=foobar
            &add_constraints.1=depends%20packagename
            &add_constraints.2=conflicts%20python%20%3C%202.5
            &remove_constraints.1=depends%20postgres

    The following errors may be raised:

    - **InvalidPackageConstraint**: The specified package constraint could not
      be parsed.
    - **InvalidConstraintType**: A constraint type other than 'depends' or
      'conflicts' was provided.
    - **EmptyPackageProfile**: Cannot remove constraints as it leaves a
      profile with no constraints. An empty package profile should be removed
      instead.

    The state of the profile will be returned.
    """

    # Note: This method is not split in "add_contraints" and
    # "remove_contraints" like most of the rest of the API since every change
    # to constraints regenerates a deb package and all associated activities
    # and is therefore quite expensive.
    requires = "ViewAccount"

    schema = Schema(
        name="EditPackageProfile",
        parameters=[
            Unicode("name", doc="The name of the package profile."),
            UnicodeTitle(
                "title", doc="The new title of the package profile.", optional=True
            ),
            List(
                "add_constraints",
                Unicode(),
                optional=True,
                doc=(
                    "List of constraints specifications to add in the form "
                    'of "depends packagename" or '
                    '"conflicts packagename < 1.0".'
                ),
            ),
            List(
                "remove_constraints",
                Unicode(),
                optional=True,
                doc=(
                    "List of constraints specifications to remove in the "
                    'form of "depends packagename" or '
                    '"conflicts packagename < 1.0".'
                ),
            ),
        ],
        errors=[
            UnknownPackageProfileNameError,
            InvalidPackageConstraintTypeError,
            InvalidPackageConstraintError,
            EmptyPackageProfileError,
            MissingParameterError,
            Unauthorized,
        ],
        doc="""
    Add or remove constraints related to a package profile. Constraints can
    be dependencies or conflicts.""",
    )

    route = ManagementRoute(factory=PackageProfilesManagement)

    def query(self, context, call):
        """
        Edit package constraints related to a package profile.

        @param context: The PackageProfile being edited.
        """
        profile = context.get_package_profile_by_name(call.args.name)
        if profile is None:
            raise UnknownPackageProfileNameError(call.args.name)

        if not any(
            (call.args.remove_constraints, call.args.add_constraints, call.args.title)
        ):
            raise MissingParameterError(
                "The request must contain at least one of the parameters "
                "add_constraints, remove_constraints or title."
            )

        context.edit(
            profile.name,
            title=call.args.title,
            add_constraints=call.args.add_constraints,
            remove_constraints=call.args.remove_constraints,
        )

        return profile.get_state()


@method
class ModifyPackageProfile(EditPackageProfile):
    """
    L{ModifyPackageProfile} is deprecated and now exists only as an alias for
    L{EditPackageProfile} to avoid breaking the API for old scripts.
    """

    schema = deepcopy(EditPackageProfile.schema)
    schema.name = "ModifyPackageProfile"
    schema.doc = (
        "THIS FUNCTION IS DEPRECATED. please use "
        + "edit-package-profile instead. "
        + schema.doc
    )


PACKAGE_PROFILE_CSV = Schema(
    name="PackageProfileCsv",
    parameters=[
        Unicode(
            "material",
            doc="Package data in the format of a CSV (as exported by Landscape)"
            " or dpkg selections.",
        )
    ],
    doc="Parse a CSV to a package profile.",
)


class DummyPackageProfile:
    """A representation of a package profile with only constraints."""

    def __init__(self) -> None:
        self.constraints: ListType[DummyPackageProfileContraint] = []

    @staticmethod
    def update_association(*args, **kwargs) -> None:
        pass

    def _add_package_constraint(self, *args: str) -> None:
        self.constraints.append(DummyPackageProfileConstraint(*args))


@dataclass
class DummyPackageProfileConstraint:
    """A representation of a single package profile constraint."""

    constraint: str
    package: str
    rule: str = ""
    version: str = ""


CREATE_CONSTRAINTS_SCHEMA = Schema(
    name="CreatePackageProfileConstraints",
    parameters=[
        Array(
            "constraints",
            ApiMapping(
                key=Unicode(),
                value=Unicode(),
                doc="The individual package constraints to create in the package profile.",
            ),
        )
    ],
    doc="Creates a constraint for a single package profile.",
)


DELETE_CONSTRAINTS_SCHEMA = Schema(
    name="DeletePackageProfileConstraints",
    parameters=[
        Array(
            "constraint_ids",
            Integer(),
            doc="The IDs of the constraints to be removed from the package profile.",
        )
    ],
    doc="Deletes constraints for a single package profile.",
)

LIST_CONSTRAINTS_SCHEMA = Schema(
    name="ListPackageProfileConstraints",
    parameters=[
        Unicode(
            "search", optional=True, doc="A term to search package constraints using."
        ),
        Unicode(
            "constraint_type",
            optional=True,
            doc="The constraint type with which to filter package constraints: 'depends' or 'conflicts'.",
        ),
    ],
    doc="Lists the package constraints for a single package profile.",
)

CREATE_SCHEMA = Schema(
    name="CreatePackageProfile",
    parameters=[
        UnicodeTitle("title", doc="The title of the package profile to create."),
        Unicode("description", doc="The description of the new profile."),
        UnicodeTitle(
            "copy_from",
            optional=True,
            doc="The name of an existing package profile to copy from.",
        ),
        Integer(
            "source_computer_id",
            optional=True,
            doc="Instance ID to be used as the basis of the new profile.",
        ),
        Unicode(
            "material",
            optional=True,
            doc="Package data in the format of 'dpkg --get-selections or Landscape-exported CSV.",
        ),
        Array(
            "constraints",
            ApiMapping(key=Unicode(), value=Unicode()),
            optional=True,
            doc="Alternative to material - individual package constraints.",
        ),
        Unicode(
            "access_group",
            optional=True,
            default=DEFAULT_ACCESS_CONTEXT_NAME,
            doc="Optional name of the access group to create the profile under.",
        ),
        Array(
            "tags",
            Unicode(),
            optional=True,
            doc="Computer tags to which to associate the profile.",
        ),
        Bool(
            "all_computers",
            optional=True,
            doc="If true, this profile will be associated with all computers.",
        ),
    ],
    doc="Creates a package profile, either specifying its constraints, or copying them from an "
    "existing package profile, instance's installed packages, 'dpkg --get-selections' output, or "
    "Landscape-exported CSV.",
)


UPDATE_CONSTRAINT_SCHEMA = Schema(
    name="UpdatePackageProfileConstraint",
    parameters=[
        Unicode("constraint", doc="The type of constraint: 'depends' or 'conflicts'."),
        Unicode("package", doc="The package to which the constraint applies."),
        Unicode(
            "rule",
            optional=True,
            doc="The comparison rule to apply against the package version: '<', '<=', '=', '>=', or '>'.",
        ),
        Unicode(
            "version",
            optional=True,
            doc="The package version against which to apply the constraint rule.",
        ),
    ],
    doc="Updates a single package constraint, generally identified by ID.",
)


UPDATE_SCHEMA = Schema(
    name="UpdatePackageProfileConstraints",
    parameters=[
        UnicodeTitle(
            "title", optional=True, doc="The new title of the package profile."
        ),
        Unicode(
            "description",
            optional=True,
            doc="The new description for the package profile.",
        ),
        Array(
            "constraints",
            ApiMapping(key=Unicode(), value=Unicode()),
            optional=True,
            doc="List of constraints to add.",
        ),
        Array(
            "tags",
            Unicode(),
            default=None,
            optional=True,
            doc="Computer tags to which to associate the profile.",
        ),
        Bool(
            "all_computers",
            optional=True,
            doc="If true, this profile will be associated with all computers.",
        ),
    ],
    doc="Update a package profile, changing its title, description, package constraints, or the "
    "instances it is associated with.",
)


@login_required
@with_schema(CREATE_SCHEMA)
async def create_package_profile_handler(
    request: Request,
    _: Principal,
    validated_data: Mapping[str, str],
) -> ApiResponse:
    """Creates a new package profile my either copying it from an existing one or creating it from
    provided parameters.
    """
    try:
        tag_names = get_tag_names(validated_data.get("tags", []))
    except InvalidTagError as e:
        return {"error": "InvalidTagError", "message": str(e)}, 400

    if validated_data["copy_from"] is not None:
        return await _copy_package_profile(
            request, validated_data=validated_data, tag_names=tag_names
        )
    else:
        return await _create_package_profile(
            request, validated_data=validated_data, tag_names=tag_names
        )

    return copy.get_state(split_constraints=True), 200


@with_management_class(PackageProfilesManagement, ["ViewAccount"])
@with_schema(GetPackageProfiles.schema)
async def list_package_profiles_handler(
    request: Request,
    _: Principal,
    management: PackageProfilesManagement,
    validated_data: Mapping[str, str],
) -> ApiResponse:
    """Responds with the details of all Package Profiles defined in the account.
    Unlike the old `GetPackageProfiles`, this one formats constraints as dicts.
    """
    profiles = management.get_package_profiles(
        names=validated_data.get("names") or None
    )

    return {
        "result": [profile.get_state(split_constraints=True) for profile in profiles]
    }, 200


@paginated
@with_management_class(PackageProfilesManagement, ["ViewAccount"])
@with_schema(LIST_CONSTRAINTS_SCHEMA)
async def list_package_profile_constraints_handler(
    unused_request: Request,
    unused_principal: Principal,
    management: PackageProfilesManagement,
    url_args: Mapping[str, str],
    validated_data: Mapping[str, str],
    search_args: Sequence[str] = (),
    offset=None,
    limit=None,
) -> ApiResponse:
    """Responds with a paginated list of package profile constraints for the given package
    profile.
    """
    profile = management.get_package_profile_by_name(url_args["package_profile_name"])

    if profile is None:
        return {
            "error": "UnknownPackageProfileNameError",
            "message": "Unknown package profile name",
        }, 404

    result = add_search_query(
        profile.constraints.find(), [PackageProfileConstraint.value], search_args
    )

    if validated_data["constraint_type"]:
        try:
            constraint_type = build_relation_type(validated_data["constraint_type"])
        except InvalidPackageConstraintTypeError:
            # Don't apply this filter if the constraint type is invalid.
            pass
        else:
            result = result.find(PackageProfileConstraint.type == constraint_type)

    count = result.count()
    result.config(limit=limit, offset=offset)

    return {
        "count": count,
        "results": [constraint.get_state() for constraint in result],
    }, 200


@with_management_class(PackageProfilesManagement, ["ViewAccount"])
@with_schema(CREATE_CONSTRAINTS_SCHEMA)
async def create_package_profile_constraints_handler(
    unused_request: Request,
    unused_principal: Principal,
    management: PackageProfilesManagement,
    url_args: Mapping[str, str],
    validated_data: Mapping[str, ListType[Mapping[str, str]]],
) -> ApiResponse:
    """Creates new package profile constraints for the given package profile."""
    profile = management.get_package_profile_by_name(url_args["package_profile_name"])

    if profile is None:
        return {
            "error": "UnknownPackageProfileNameError",
            "message": "Unknown package profile name",
        }, 404

    if not validated_data["constraints"]:
        return {
            "error": "NoPackageProfileConstraintsError",
            "message": "No constraints provided to create",
        }, 400

    try:
        result = profile.modify_package_constraints(
            add_constraints=validated_data["constraints"],
            return_as_state=True,
        )
    except (
        EmptyPackageProfileError,
        InvalidPackageConstraintError,
        InvalidPackageConstraintTypeError,
    ) as e:
        return {"error": e.__class__.__name__, "message": str(e)}, 400

    return {"results": result}, 201


@with_management_class(PackageProfilesManagement, ["ViewAccount"])
@with_schema(DELETE_CONSTRAINTS_SCHEMA)
async def delete_package_profile_constraints_handler(
    unused_request: Request,
    unused_principal: Principal,
    management: PackageProfilesManagement,
    url_args: Mapping[str, str],
    validated_data: Mapping[str, ListType[int]],
) -> ApiResponse:
    """Deletes the indicated package profile constraints for the given package profile."""
    profile = management.get_package_profile_by_name(url_args["package_profile_name"])

    if profile is None:
        return {
            "error": "UnknownPackageProfileNameError",
            "message": "Unknown package profile name",
        }, 404

    if not validated_data["constraint_ids"]:
        return {
            "error": "NoPackageProfileConstraintsError",
            "message": "No constraints provided to delete",
        }, 400

    constraints = profile.constraints.find(
        PackageProfileConstraint.id.is_in(validated_data["constraint_ids"])
    )

    try:
        result = profile.modify_package_constraints(
            remove_constraints=[constraint.parse() for constraint in constraints]
        )
    except EmptyPackageProfileError as e:
        return {"error": "EmptyPackageProfileError", "message": str(e)}, 400

    return EMPTY_BODY, 204


@with_management_class(PackageProfilesManagement, ["ViewAccount"])
@with_schema(UPDATE_CONSTRAINT_SCHEMA)
async def update_package_profile_constraint_handler(
    unused_request: Request,
    unused_principal: Principal,
    management: PackageProfilesManagement,
    url_args: Mapping[str, Union[str, int]],
    validated_data: Mapping[str, str],
) -> ApiResponse:
    """Deletes the indicated package profile constraints for the given package profile."""
    profile = management.get_package_profile_by_name(url_args["package_profile_name"])

    if profile is None:
        return {
            "error": "UnknownPackageProfileNameError",
            "message": "Unknown package profile name",
        }, 404

    constraint = profile.constraints.find(
        PackageProfileConstraint.id == url_args["id"]
    ).one()

    if constraint is None:
        return {
            "error": "UnknownPackageProfileConstraint",
            "message": "Unknown package profile constraint or constraint does not belong to this profile.",
        }, 404

    try:
        parsed = parse_constraint(list(validated_data.items()))
    except (InvalidPackageConstraintError, InvalidPackageConstraintTypeError) as e:
        return {"error": e.__class__.__name__, "message": str(e)}, 400

    constraint.type = build_relation_type(validated_data["constraint"])
    constraint.value = build_relation_value([parsed[1]])

    return constraint.get_state(), 200


@with_computer(Schema(), "ViewComputer")
async def get_computer_package_profile_handler(
    request: Request,
    principal: Principal,
    computer: Computer,
    selection: ComputerSelection,
    validated_data: Mapping,
    url_args: Mapping[str, int],
) -> ApiResponse:
    """Responds with a dict-serialized package profile, built from the installed packages of the
    provided computer.
    """
    try:
        profile = create_profile_from_computer_selection(DummyPackageProfile, selection)
    except NoFoundPackagesError as e:
        return {"error": "NoFoundPackagesError", "message": str(e)}, 400

    return {"result": [asdict(constraint) for constraint in profile.constraints]}, 200


@login_required
@with_schema(PACKAGE_PROFILE_CSV)
async def parse_package_profile_material_handler(
    request: Request, _: Principal, validated_data: Mapping[str, str]
) -> ApiResponse:
    """Attempts to parse a CSV-formatted package profile, producing a list of constraints if
    successful or an explanatory error if not.
    """
    # Twisted's Requests double-escape control characters. Luckily we already know we have valid
    # UTF-8 thanks to `with_schema`.
    material = validated_data["material"].encode().decode("unicode-escape")

    try:
        profile = parse_package_profile_material(DummyPackageProfile, material)
    except InvalidPackageProfileMaterialError as e:
        return {"error": "InvalidPackageProfileMaterialError", "message": str(e)}, 400

    return {"result": [asdict(constraint) for constraint in profile.constraints]}, 200


@with_management_class(PackageProfilesManagement, ["ViewAccount"])
@with_schema(UPDATE_SCHEMA)
async def update_package_profile_handler(
    unused_request: Request,
    unused_principal: Principal,
    management: PackageProfilesManagement,
    validated_data: Mapping[str, str],
    url_args: Mapping[str, str],
) -> ApiResponse:
    """Updates the constraints associated with a package profile."""
    profile = management.get_package_profile_by_name(url_args["package_profile_name"])

    if profile is None:
        return {
            "error": "UnknownPackageProfileNameError",
            "message": "Unknown package profile name.",
        }, 404

    if not any(
        (
            validated_data["constraints"],
            validated_data["title"],
            validated_data["description"],
            validated_data["tags"] is not None,
            validated_data["all_computers"] is not None,
        )
    ):
        return {
            "error": "MissingParameterError",
            "message": "The request must contain at least one of constraints, title, tags, "
            "all_computers, or description.",
        }, 400

    try:
        tag_names = get_tag_names(validated_data.get("tags", []))
    except InvalidTagError as e:
        return {"error": "InvalidTagError", "message": str(e)}, 400

    try:
        management.edit(
            profile.name,
            title=validated_data["title"],
            add_constraints=validated_data["constraints"],
            replace_constraints=True,
        )
    except (
        InvalidPackageConstraintTypeError,
        InvalidPackageConstraintError,
        PackageProfileError,
        Unauthorized,
    ) as e:
        return {"error": e.__class__.__name__, "message": str(e)}, 400

    all_computers = validated_data.get("all_computers")

    if tag_names:
        tags = get_tags(tag_names)
    else:
        tags = []

    if tags or all_computers is not None:
        profile.update_association(tags, all_computers)

    return profile.get_state(split_constraints=True), 200


@with_management_class(PackageProfilesManagement, ["ViewAccount"])
async def _copy_package_profile(
    unused_request: Request,
    unused_prinicial: Principal,
    management: PackageProfilesManagement,
    validated_data: Mapping[str, str],
    tag_names: ListType[str] = [],
) -> ApiResponse:
    """Creates a new package profile by copying it from an existing one.
    This isn't a 'real' API handler - it's a helper used by `create_package_profile_handler`.
    """
    try:
        copy = management.copy_package_profile(
            # Can assume presence of 'copy_from', as this function wouldn't get called otherwise.
            validated_data["copy_from"],
            name=validated_data["title"],
            title=validated_data["title"],
            description=validated_data["description"],
            access_context_name=validated_data["access_group"],
            copy_association=True,
        )
    except (
        UnknownPackageProfileNameError,
        DuplicatePackageProfileError,
        UnknownAccessContextError,
        Unauthorized,
    ) as e:
        return {"error": e.__class__.__name__, "message": str(e)}, 400

    all_computers = validated_data.get("all_computers")

    if tag_names:
        tags = get_tags(tag_names)
    else:
        tags = []

    if tags or all_computers is not None:
        copy.update_association(tags, all_computers)

    return copy.get_state(split_constraints=True), 200


@with_management_class(PackageProfilesManagement, ["ViewAccount"])
async def _create_package_profile(
    unused_request: Request,
    principal: Principal,
    management: PackageProfilesManagement,
    validated_data: Mapping[str, str],
    tag_names: ListType[str] = [],
) -> ApiResponse:
    """Creates a new package profile using the provided parameters.
    This isn't a 'real' API handler - it's a helper used by `create_package_profile_handler`.
    """
    try:
        profile = management.create_package_profile(
            validated_data["title"],
            validated_data["description"],
            material=validated_data["material"],
            computer=validated_data["source_computer_id"],
            access_context_name=validated_data["access_group"],
        )
        profile.modify_package_constraints(
            add_constraints=validated_data["constraints"]
        )
    except (
        EmptyPackageProfileError,
        InvalidPackageProfileNameError,
        InvalidPackageProfileMaterialError,
        NoFoundPackagesError,
        UnknownAccessContextError,
        Unauthorized,
    ) as e:
        return {"error": e.__class__.__name__, "message": str(e)}, 400

    profile.on_create(principal.person)

    all_computers = validated_data.get("all_computers")

    if tag_names:
        tags = get_tags(tag_names)
    else:
        tags = []

    if tags or all_computers is not None:
        profile.update_association(tags, all_computers)

    return profile.get_state(split_constraints=True), 200
